from infrahub.core.constants import InfrahubKind
from infrahub.trigger.models import BuiltinTriggerDefinition, EventTrigger, ExecuteWorkflow
from infrahub.workflows.catalogue import WEBHOOK_CONFIGURE_ONE, WEBHOOK_DELETE_AUTOMATION

TRIGGER_WEBHOOK_SETUP_UPDATE = BuiltinTriggerDefinition(
    name="webhook-configure-one",
    trigger=EventTrigger(
        events={"infrahub.node.created", "infrahub.node.updated"},
        match={
            "infrahub.node.kind": [InfrahubKind.CUSTOMWEBHOOK, InfrahubKind.STANDARDWEBHOOK],
        },
    ),
    actions=[
        ExecuteWorkflow(
            workflow=WEBHOOK_CONFIGURE_ONE,
            parameters={
                "webhook_name": "{{ event.payload['data']['changelog']['display_label'] }}",
                "event_data": {
                    "__prefect_kind": "json",
                    "value": {"__prefect_kind": "jinja", "template": "{{ event.payload['data'] | tojson }}"},
                },
            },
        ),
    ],
)

TRIGGER_WEBHOOK_DELETE = BuiltinTriggerDefinition(
    name="webhook-delete",
    trigger=EventTrigger(
        events={"infrahub.node.deleted"},
        match={
            "infrahub.node.kind": [InfrahubKind.CUSTOMWEBHOOK, InfrahubKind.STANDARDWEBHOOK],
        },
    ),
    actions=[
        ExecuteWorkflow(
            workflow=WEBHOOK_DELETE_AUTOMATION,
            parameters={
                "webhook_id": "{{ event.payload['data']['node_id'] }}",
                "webhook_name": "{{ event.payload['data']['changelog']['display_label'] }}",
            },
        ),
    ],
)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from infrahub_sdk.graphql import Query
from prefect.events.schemas.automations import Automation  # noqa: TC002
from pydantic import BaseModel, ConfigDict, Field, computed_field
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import RelationshipCardinality
from infrahub.core.schema import AttributeSchema, NodeSchema  # noqa: TC001
from infrahub.core.schema.schema_branch_computed import (  # noqa: TC001
    ComputedAttributeTarget,
    ComputedAttributeTriggerNode,
    PythonDefinition,
)
from infrahub.events import NodeCreatedEvent, NodeUpdatedEvent
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer  # noqa: TC001
from infrahub.trigger.constants import NAME_SEPARATOR
from infrahub.trigger.models import (
    EventTrigger,
    ExecuteWorkflow,
    TriggerBranchDefinition,
    TriggerType,
)
from infrahub.workflows.catalogue import (
    COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
    COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
    QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS,
)

if TYPE_CHECKING:
    from uuid import UUID

    from infrahub.git.models import RepositoryData


class ComputedAttributeAutomations(BaseModel):
    data: dict[str, dict[str, Automation]] = Field(default_factory=lambda: defaultdict(dict))  # type: ignore[arg-type]

    @classmethod
    def from_prefect(cls, automations: list[Automation], prefix: str = "") -> Self:
        obj = cls()
        for automation in automations:
            if not automation.name.startswith(prefix):
                continue

            name_split = automation.name.split(NAME_SEPARATOR)
            if len(name_split) != 3:
                continue

            scope = name_split[1]
            identifier = name_split[2]

            obj.data[identifier][scope] = automation

        return obj

    def get(self, identifier: str, scope: str) -> Automation:
        if identifier in self.data and scope in self.data[identifier]:
            return self.data[identifier][scope]
        raise KeyError(f"Unable to find an automation for {identifier} {scope}")

    def has(self, identifier: str, scope: str) -> bool:
        if identifier in self.data and scope in self.data[identifier]:
            return True
        return False

    @property
    def all_automation_ids(self) -> list[UUID]:
        automation_ids: list[UUID] = []
        for identifier in self.data.values():
            for automation in identifier.values():
                automation_ids.append(automation.id)
        return automation_ids


class PythonTransformComputedAttribute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    repository_id: str
    repository_name: str
    repository_kind: str
    query_name: str
    query_analyzer: InfrahubGraphQLQueryAnalyzer
    computed_attribute: PythonDefinition
    default_schema: bool
    branch_name: str
    branch_commit: dict[str, str] = field(default_factory=dict)

    @computed_field
    def repository_commit(self) -> str:
        return self.branch_commit[self.branch_name]

    def populate_branch_commit(self, repository_data: RepositoryData | None = None) -> None:
        if repository_data:
            for branch, commit in repository_data.branches.items():
                self.branch_commit[branch] = commit

    def get_altered_branches(self) -> list[str]:
        if registry.default_branch in self.branch_commit:
            default_branch_commit = self.branch_commit[registry.default_branch]
            return [
                branch_name for branch_name, commit in self.branch_commit.items() if commit != default_branch_commit
            ]
        return list(self.branch_commit.keys())


@dataclass
class PythonTransformTarget:
    kind: str
    object_id: str


class ComputedAttrJinja2TriggerDefinition(TriggerBranchDefinition):
    type: TriggerType = TriggerType.COMPUTED_ATTR_JINJA2
    computed_attribute: ComputedAttributeTarget
    template_hash: str
    trigger_kind: str

    @property
    def targets_self(self) -> bool:
        """Determine if the specific trigger definition targets the actual node kind of the computed attribute."""
        return self.trigger_kind == self.computed_attribute.kind

    def get_description(self) -> str:
        return f"{super().get_description()} | hash:{self.template_hash}"

    @classmethod
    def from_computed_attribute(
        cls,
        branch: str,
        computed_attribute: ComputedAttributeTarget,
        trigger_node: ComputedAttributeTriggerNode,
        branches_out_of_scope: list[str] | None = None,
    ) -> Self:
        """
        This function is used to create a trigger definition for a computed attribute of type Jinja2.
        """
        event_trigger = EventTrigger()
        event_trigger.events.add(NodeUpdatedEvent.event_name)
        if computed_attribute.attribute.optional:
            # If the computed attribute not optional it means that Infrahub will have assigned the value during
            # the creation of the node so we don't need to match on node creation events as it would only add
            # extra work that doesn't accomplish anything. For this reason the filter should only match on the
            # node creation events if the attribute is optional.
            event_trigger.events.add(NodeCreatedEvent.event_name)

        if (
            computed_attribute.attribute.computed_attribute
            and computed_attribute.attribute.computed_attribute.jinja2_template is None
        ) or not computed_attribute.attribute.computed_attribute:
            raise ValueError("Jinja2 template is required for computed attribute")

        template_hash = computed_attribute.attribute.computed_attribute.get_hash()

        event_trigger.match = {"infrahub.node.kind": trigger_node.kind}
        if branches_out_of_scope:
            event_trigger.match["infrahub.branch.name"] = [f"!{branch}" for branch in branches_out_of_scope]
        elif not branches_out_of_scope and branch != registry.default_branch:
            event_trigger.match["infrahub.branch.name"] = branch

        event_trigger.match_related = {
            "prefect.resource.role": ["infrahub.node.attribute_update", "infrahub.node.relationship_update"],
            "infrahub.field.name": trigger_node.fields,
        }

        workflow = ExecuteWorkflow(
            workflow=COMPUTED_ATTRIBUTE_PROCESS_JINJA2,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                "object_id": "{{ event.resource['infrahub.node.id'] }}",
                "computed_attribute_name": computed_attribute.attribute.name,
                "computed_attribute_kind": computed_attribute.kind,
                "updated_fields": {
                    "__prefect_kind": "json",
                    "value": {
                        "__prefect_kind": "jinja",
                        "template": "{{ event.payload['data']['fields'] | tojson }}",
                    },
                },
                "context": {
                    "__prefect_kind": "json",
                    "value": {
                        "__prefect_kind": "jinja",
                        "template": "{{ event.payload['context'] | tojson }}",
                    },
                },
            },
        )

        definition = cls(
            name=f"{computed_attribute.key_name}{NAME_SEPARATOR}kind{NAME_SEPARATOR}{trigger_node.kind}",
            template_hash=template_hash,
            trigger_kind=trigger_node.kind,
            branch=branch,
            computed_attribute=computed_attribute,
            trigger=event_trigger,
            actions=[workflow],
        )

        return definition


class ComputedAttrPythonTriggerDefinition(TriggerBranchDefinition):
    type: TriggerType = TriggerType.COMPUTED_ATTR_PYTHON
    computed_attribute: PythonTransformComputedAttribute

    @classmethod
    def from_object(
        cls,
        branch: str,
        computed_attribute: PythonTransformComputedAttribute,
        branches_out_of_scope: list[str] | None = None,
    ) -> Self:
        # scope = registry.default_branch

        event_trigger = EventTrigger()
        event_trigger.events.update({NodeCreatedEvent.event_name, NodeUpdatedEvent.event_name})
        event_trigger.match = {
            "infrahub.node.kind": [computed_attribute.computed_attribute.kind],
        }

        if branches_out_of_scope:
            event_trigger.match["infrahub.branch.name"] = [f"!{branch}" for branch in branches_out_of_scope]
        elif not branches_out_of_scope and branch != registry.default_branch:
            event_trigger.match["infrahub.branch.name"] = branch

        update_fields = computed_attribute.query_analyzer.query_report.fields_by_kind(
            kind=computed_attribute.computed_attribute.kind
        )
        event_trigger.match_related = {
            "prefect.resource.role": ["infrahub.node.attribute_update", "infrahub.node.relationship_update"],
        }

        if update_fields and "display_label" not in update_fields:
            # The GraphQLQuery analyzer doesn't yet support figuring out which updates would match the "display label"
            # of a query. Because of this we temporarily match any field if the display_label is part of the computed
            # attribute query
            event_trigger.match_related["infrahub.field.name"] = update_fields

        definition = cls(
            name=computed_attribute.computed_attribute.key_name,
            branch=branch,
            computed_attribute=computed_attribute,
            trigger=event_trigger,
            actions=[
                ExecuteWorkflow(
                    workflow=COMPUTED_ATTRIBUTE_PROCESS_TRANSFORM,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                        "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                        "object_id": "{{ event.resource['infrahub.node.id'] }}",
                        "computed_attribute_name": computed_attribute.computed_attribute.attribute.name,
                        "computed_attribute_kind": computed_attribute.computed_attribute.kind,
                        "context": {
                            "__prefect_kind": "json",
                            "value": {
                                "__prefect_kind": "jinja",
                                "template": "{{ event.payload['context'] | tojson }}",
                            },
                        },
                    },
                ),
            ],
        )

        return definition


class ComputedAttrPythonQueryTriggerDefinition(TriggerBranchDefinition):
    type: TriggerType = TriggerType.COMPUTED_ATTR_PYTHON_QUERY

    @classmethod
    def from_object(
        cls,
        branch: str,
        kind: str,
        computed_attribute: PythonTransformComputedAttribute,
        branches_out_of_scope: list[str] | None = None,
    ) -> Self:
        # Only matching on node updated events, before nodes are created they won't be a member of the GraphQL query
        # group regardless so it doesn't make sense to trigger the query on node creation. For the initial object
        # where the computed attribute belongs that to will need to be created first which will trigger its own initial
        # process to populated the original members of the GraphQL query group
        event_trigger = EventTrigger(events={NodeUpdatedEvent.event_name})
        event_trigger.match = {
            "infrahub.node.kind": kind,
        }

        update_fields = computed_attribute.query_analyzer.query_report.fields_by_kind(kind=kind)
        event_trigger.match_related = {
            "prefect.resource.role": ["infrahub.node.attribute_update", "infrahub.node.relationship_update"],
        }

        if update_fields and "display_label" not in update_fields:
            # The GraphQLQuery analyzer doesn't yet support figuring out which updates would match the "display label"
            # of a query. Because of this we temporarily match any field if the display_label is part of the computed
            # attribute query
            event_trigger.match_related["infrahub.field.name"] = update_fields

        if branches_out_of_scope:
            event_trigger.match["infrahub.branch.name"] = [f"!{branch}" for branch in branches_out_of_scope]
        elif not branches_out_of_scope and branch != registry.default_branch:
            event_trigger.match["infrahub.branch.name"] = branch

        definition = cls(
            name=f"{computed_attribute.computed_attribute.key_name}{NAME_SEPARATOR}kind{NAME_SEPARATOR}{kind}",
            branch=branch,
            trigger=event_trigger,
            actions=[
                ExecuteWorkflow(
                    workflow=QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                        "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                        "object_id": "{{ event.resource['infrahub.node.id'] }}",
                        "context": {
                            "__prefect_kind": "json",
                            "value": {
                                "__prefect_kind": "jinja",
                                "template": "{{ event.payload['context'] | tojson }}",
                            },
                        },
                    },
                ),
            ],
        )

        return definition


class ComputedAttrJinja2GraphQLResponse(BaseModel):
    node_id: str
    computed_attribute_value: str | None
    variables: dict[str, Any] = Field(default_factory=dict)


class ComputedAttrJinja2GraphQL(BaseModel):
    node_schema: NodeSchema = Field(..., description="The node kind where the computed attribute is defined")
    attribute_schema: AttributeSchema = Field(..., description="The computed attribute")
    variables: list[str] = Field(..., description="The list of variable names used within the computed attribute")

    def render_graphql_query(self, query_filter: str, filter_id: str) -> str:
        query_fields = self.query_fields
        query_fields["id"] = None
        query_fields[self.attribute_schema.name] = {"value": None}
        query = Query(
            name="ComputedAttributeFilter",
            query={
                self.node_schema.kind: {
                    "@filters": {query_filter: filter_id},
                    "edges": {"node": query_fields},
                }
            },
        )

        return query.render()

    @property
    def query_fields(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for variable in self.variables:
            field_name, remainder = variable.split("__", maxsplit=1)
            if field_name in self.node_schema.attribute_names:
                output[field_name] = {remainder: None}
            elif field_name in self.node_schema.relationship_names:
                related_attribute, related_value = remainder.split("__", maxsplit=1)
                relationship = self.node_schema.get_relationship(name=field_name)
                if relationship.cardinality == RelationshipCardinality.ONE:
                    if field_name not in output:
                        output[field_name] = {"node": {}}
                    output[field_name]["node"][related_attribute] = {related_value: None}
        return output

    def parse_response(self, response: dict[str, Any]) -> list[ComputedAttrJinja2GraphQLResponse]:
        rendered_response: list[ComputedAttrJinja2GraphQLResponse] = []
        if kind_payload := response.get(self.node_schema.kind):
            edges = kind_payload.get("edges", [])
            for node in edges:
                if node_response := self.to_node_response(node_dict=node):
                    rendered_response.append(node_response)
        return rendered_response

    def to_node_response(self, node_dict: dict[str, Any]) -> ComputedAttrJinja2GraphQLResponse | None:
        if node := node_dict.get("node"):
            node_id = node.get("id")
        else:
            return None

        computed_attribute = node.get(self.attribute_schema.name, {}).get("value")
        response = ComputedAttrJinja2GraphQLResponse(node_id=node_id, computed_attribute_value=computed_attribute)
        for variable in self.variables:
            field_name, remainder = variable.split("__", maxsplit=1)
            response.variables[variable] = None
            if field_content := node.get(field_name):
                if field_name in self.node_schema.attribute_names:
                    response.variables[variable] = field_content.get(remainder)
                elif field_name in self.node_schema.relationship_names:
                    relationship = self.node_schema.get_relationship(name=field_name)
                    if relationship.cardinality == RelationshipCardinality.ONE:
                        related_attribute, related_value = remainder.split("__", maxsplit=1)
                        node_content = field_content.get("node") or {}
                        related_attribute_content = node_content.get(related_attribute) or {}
                        response.variables[variable] = related_attribute_content.get(related_value)

        return response



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Self, cast

from graphene import InputObjectType, Mutation

from infrahub.core.manager import NodeManager
from infrahub.core.schema import NodeSchema
from infrahub.database import retry_db_transaction
from infrahub.events.utils import get_all_infrahub_node_kind_events
from infrahub.exceptions import ValidationError
from infrahub.log import get_logger

from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.protocols import CoreWebhook
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

log = get_logger()


@dataclass
class StringValue:
    value: str


@dataclass
class WebhookCreate:
    event_type: StringValue | None = field(default=None)
    node_kind: StringValue | None = field(default=None)


class WebhookUpdate(WebhookCreate):
    id: str | None = field(default=None)
    hfid: list[str] | None = field(default=None)


class InfrahubWebhookMutation(InfrahubMutationMixin, Mutation):
    _meta: InfrahubMutationOptions

    @classmethod
    def __init_subclass_with_meta__(
        cls, schema: NodeSchema | None = None, _meta: InfrahubMutationOptions | None = None, **options: dict[str, Any]
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        input_data = cast("WebhookCreate", data)

        _validate_input(graphql_context=graphql_context, branch=branch, input_data=input_data)

        obj, result = await super().mutate_create(
            info=info, data=data, branch=branch, database=database, override_data=override_data
        )

        return obj, result

    @classmethod
    @retry_db_transaction(name="object_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        db = database or graphql_context.db

        input_data = cast("WebhookUpdate", data)

        _validate_input(graphql_context=graphql_context, branch=branch, input_data=input_data)

        obj = node or await NodeManager.find_object(
            db=db,
            kind=cls._meta.active_schema.kind,
            id=input_data.id,
            hfid=input_data.hfid,
            branch=branch,
        )

        webhook = cast("CoreWebhook", obj)

        event_type = input_data.event_type.value if input_data.event_type else webhook.event_type.value.value
        node_kind = input_data.node_kind.value if input_data.node_kind else webhook.node_kind.value

        if event_type and node_kind:
            updated_data = WebhookUpdate(
                event_type=StringValue(value=event_type), node_kind=StringValue(value=node_kind)
            )
            _validate_input(graphql_context=graphql_context, branch=branch, input_data=updated_data)

        try:
            obj, result = await cls._call_mutate_update(info=info, data=data, db=db, branch=branch, obj=obj)
        except ValidationError as exc:
            raise ValueError(str(exc)) from exc

        return obj, result


def _validate_input(graphql_context: GraphqlContext, branch: Branch, input_data: WebhookCreate) -> None:
    if input_data.node_kind and input_data.node_kind.value:
        # Validate that the requested node_kind exists, will raise an error if not
        graphql_context.db.schema.get(name=input_data.node_kind.value, branch=branch, duplicate=False)

        if input_data.event_type:
            # If the event type is not set then all events are applicable, this will mean that some events
            # would be filtered out, as they won't match the node.
            if input_data.event_type.value not in get_all_infrahub_node_kind_events():
                raise ValidationError(
                    input_value=f"Defining a node_kind is not valid for {input_data.event_type.value} events"
                )



from __future__ import annotations

from collections.abc import Iterable
from copy import copy, deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

from ..constants import InfrahubClientMode
from ..exceptions import FeatureNotSupportedError, NodeNotFoundError, ResourceNotDefinedError, SchemaNotFoundError
from ..file_handler import FileHandler, FileHandlerBase, FileHandlerSync, PreparedFile
from ..graphql import Mutation, Query
from ..schema import (
    GenericSchemaAPI,
    ProfileSchemaAPI,
    RelationshipCardinality,
    RelationshipKind,
    RelationshipSchemaAPI,
    TemplateSchemaAPI,
)
from ..utils import compare_lists, generate_short_id
from .attribute import Attribute
from .constants import (
    ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    FILE_DOWNLOAD_FEATURE_NOT_SUPPORTED_MESSAGE,
    PROPERTIES_OBJECT,
)
from .metadata import NodeMetadata
from .related_node import RelatedNode, RelatedNodeBase, RelatedNodeSync
from .relationship import RelationshipManager, RelationshipManagerBase, RelationshipManagerSync

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..client import InfrahubClient, InfrahubClientSync
    from ..context import RequestContext
    from ..schema import MainSchemaTypesAPI
    from ..types import Order


class InfrahubNodeBase:
    """Base class for InfrahubNode and InfrahubNodeSync"""

    def __init__(self, schema: MainSchemaTypesAPI, branch: str, data: dict | None = None) -> None:
        """
        Args:
            schema: The schema of the node.
            branch: The branch where the node resides.
            data: Optional data to initialize the node.
        """
        self._schema = schema
        self._data = data
        self._branch = branch
        self._existing: bool = True
        self._attribute_data: dict[str, Attribute] = {}
        self._metadata: NodeMetadata | None = None

        # Generate a unique ID only to be used inside the SDK
        # The format if this ID is purposely different from the ID used by the API
        # This is done to avoid confusion and potential conflicts between the IDs
        self._internal_id = generate_short_id()

        self.id = data.get("id", None) if isinstance(data, dict) else None
        self.display_label: str | None = data.get("display_label", None) if isinstance(data, dict) else None
        self.typename: str | None = data.get("__typename", schema.kind) if isinstance(data, dict) else schema.kind

        self._attributes = [item.name for item in self._schema.attributes]
        self._relationships = [item.name for item in self._schema.relationships]

        # GenericSchemaAPI doesn't have inherit_from
        inherit_from: list[str] = getattr(schema, "inherit_from", None) or []
        self._artifact_support = "CoreArtifactTarget" in inherit_from
        self._file_object_support = "CoreFileObject" in inherit_from
        self._artifact_definition_support = schema.kind == "CoreArtifactDefinition"

        self._file_content: bytes | Path | BinaryIO | None = None
        self._file_name: str | None = None

        # Check if this node is hierarchical (supports parent/children and ancestors/descendants)
        if not isinstance(schema, (ProfileSchemaAPI, GenericSchemaAPI, TemplateSchemaAPI)):
            self._hierarchy_support = getattr(schema, "hierarchy", None) is not None
        else:
            self._hierarchy_support = False

        if not self.id:
            self._existing = False

        self._init_attributes(data)
        self._init_relationships(data)

    def get_branch(self) -> str:
        return self._branch

    def get_path_value(self, path: str) -> Any:
        path_parts = path.split("__")
        return_value = None

        # Manage relationship value lookup
        if path_parts[0] in self._schema.relationship_names:
            related_node = getattr(self, path_parts[0], None)
            if not related_node:
                return None

            try:
                peer = related_node.get()
            except (NodeNotFoundError, ValueError):
                # This can happen while batch creating nodes, the lookup won't work as the store is not populated
                # If so we cannot complete the HFID computation as we cannot access the related node attribute value
                return None

            if attribute_piece := path_parts[1] if len(path_parts) > 1 else None:
                related_node_attribute = getattr(peer, attribute_piece, None)
            else:
                return peer.hfid or peer.id

            if property_piece := path_parts[2] if len(path_parts) > 2 else None:
                return_value = getattr(related_node_attribute, property_piece, None)
            else:
                return_value = related_node_attribute

        # Manage attribute value lookup
        if path_parts[0] in self._schema.attribute_names:
            attribute = getattr(self, path_parts[0], None)
            if property_piece := path_parts[1] if len(path_parts) > 1 else None:
                return_value = getattr(attribute, property_piece, None)
            else:
                return_value = attribute

        return return_value

    def get_human_friendly_id(self) -> list[str] | None:
        if not hasattr(self._schema, "human_friendly_id"):
            return None

        if not self._schema.human_friendly_id:
            return None

        # If an HFID component is missing we assume that it is invalid and not usable for this node
        hfid_components = [self.get_path_value(path=item) for item in self._schema.human_friendly_id]
        if None in hfid_components:
            return None
        return [str(hfid) for hfid in hfid_components]

    def get_human_friendly_id_as_string(self, include_kind: bool = False) -> str | None:
        hfid = self.get_human_friendly_id()
        if not hfid:
            return None
        if include_kind:
            hfid = [self.get_kind(), *hfid]
        return "__".join(hfid)

    @property
    def hfid(self) -> list[str] | None:
        return self.get_human_friendly_id()

    @property
    def hfid_str(self) -> str | None:
        return self.get_human_friendly_id_as_string(include_kind=True)

    def get_node_metadata(self) -> NodeMetadata | None:
        """Returns the node metadata (created_at, created_by, updated_at, updated_by) if fetched."""
        return self._metadata

    def _init_attributes(self, data: dict | None = None) -> None:
        for attr_schema in self._schema.attributes:
            attr_data = data.get(attr_schema.name, None) if isinstance(data, dict) else None
            self._attribute_data[attr_schema.name] = Attribute(
                name=attr_schema.name, schema=attr_schema, data=attr_data
            )

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for attributes that exist or revert to normal behaviour"""
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            self._attribute_data[name].value = value
            return

        super().__setattr__(name, value)

    def _get_request_context(self, request_context: RequestContext | None = None) -> dict[str, Any] | None:
        if request_context:
            return request_context.model_dump(exclude_none=True)

        client: InfrahubClient | InfrahubClientSync | None = getattr(self, "_client", None)
        if not client or not client.request_context:
            return None

        return client.request_context.model_dump(exclude_none=True)

    def _init_relationships(self, data: dict | None = None) -> None:
        pass

    def __repr__(self) -> str:
        if self.display_label:
            return self.display_label
        if not self._existing:
            return f"{self._schema.kind} ({self.id})[NEW]"

        return f"{self._schema.kind} ({self.id}) "

    def get_kind(self) -> str:
        return self._schema.kind

    def get_all_kinds(self) -> list[str]:
        if inherit_from := getattr(self._schema, "inherit_from", None):
            return [self._schema.kind, *inherit_from]
        return [self._schema.kind]

    def is_ip_prefix(self) -> bool:
        builtin_ipprefix_kind = "BuiltinIPPrefix"
        return self.get_kind() == builtin_ipprefix_kind or builtin_ipprefix_kind in self._schema.inherit_from  # type: ignore[union-attr]

    def is_ip_address(self) -> bool:
        builtin_ipaddress_kind = "BuiltinIPAddress"
        return self.get_kind() == builtin_ipaddress_kind or builtin_ipaddress_kind in self._schema.inherit_from  # type: ignore[union-attr]

    def is_resource_pool(self) -> bool:
        return hasattr(self._schema, "inherit_from") and "CoreResourcePool" in self._schema.inherit_from  # type: ignore[union-attr]

    def is_file_object(self) -> bool:
        """Check if this node inherits from CoreFileObject and supports file uploads."""
        return self._file_object_support

    def upload_from_path(self, path: Path) -> None:
        """Set a file from disk to be uploaded when saving this FileObject node.

        The file will be streamed during upload, avoiding loading the entire file into memory.

        Args:
            path: Path to the file on disk.

        Raises:
            FeatureNotSupportedError: If this node doesn't inherit from CoreFileObject.

        Example:
            node.upload_from_path(path=Path("/path/to/large_file.pdf"))
        """
        if not self._file_object_support:
            raise FeatureNotSupportedError(
                f"File upload is not supported for {self._schema.kind}. Only nodes inheriting from CoreFileObject support file uploads."
            )
        self._file_content = path
        self._file_name = path.name

    def upload_from_bytes(self, content: bytes | BinaryIO, name: str) -> None:
        """Set content to be uploaded when saving this FileObject node.

        The content can be provided as bytes or a file-like object.
        Using BinaryIO is recommended for large content to stream during upload.

        Args:
            content: The file content as bytes or a file-like object.
            name: The filename to use for the uploaded file.

        Raises:
            FeatureNotSupportedError: If this node doesn't inherit from CoreFileObject.

        Examples:
            >>> # Using bytes (for small files)
            >>> node.upload_from_bytes(content=b"file content", name="example.txt")

            >>> # Using file-like object (for large files)
            >>> with open("/path/to/file.bin", "rb") as f:
            ...     node.upload_from_bytes(content=f, name="file.bin")
        """
        if not self._file_object_support:
            raise FeatureNotSupportedError(
                f"File upload is not supported for {self._schema.kind}. Only nodes inheriting from CoreFileObject support file uploads."
            )
        self._file_content = content
        self._file_name = name

    def clear_file(self) -> None:
        """Clear any pending file content."""
        self._file_content = None
        self._file_name = None

    async def _get_file_for_upload(self) -> PreparedFile:
        """Get the file content as a file-like object for upload (async version)."""
        return await FileHandlerBase.prepare_upload(content=self._file_content, name=self._file_name)

    def _get_file_for_upload_sync(self) -> PreparedFile:
        """Get the file content as a file-like object for upload (sync version)."""
        return FileHandlerBase.prepare_upload_sync(content=self._file_content, name=self._file_name)

    def get_raw_graphql_data(self) -> dict | None:
        return self._data

    def _generate_input_data(  # noqa: C901
        self,
        exclude_unmodified: bool = False,
        exclude_hfid: bool = False,
        request_context: RequestContext | None = None,
    ) -> dict[str, dict]:
        """Generate a dictionary that represent the input data required by a mutation.

        Returns:
            dict[str, Dict]: Representation of an input data in dict format
        """

        data: dict[str, Any] = {}
        variables: dict[str, Any] = {}

        for item_name in self._attributes:
            attr: Attribute = getattr(self, item_name)
            if attr._schema.read_only:
                continue
            graphql_payload = attr._generate_input_data()
            if graphql_payload.payload:
                data[item_name] = graphql_payload.payload
            if graphql_payload.variables:
                variables.update(graphql_payload.variables)

        for item_name in self._relationships:
            allocate_from_pool = False
            rel_schema = self._schema.get_relationship(name=item_name)
            if not rel_schema or rel_schema.read_only:
                continue

            rel: RelatedNodeBase | RelationshipManagerBase = getattr(self, item_name)

            if rel_schema.cardinality == RelationshipCardinality.ONE and rel_schema.optional and not rel.initialized:
                # Only include None for existing nodes to allow clearing relationships
                # For new nodes, omit the field to allow object template defaults to be applied
                if self._existing:
                    data[item_name] = None
                continue

            if rel is None or not rel.initialized:
                continue

            if isinstance(rel, (RelatedNode, RelatedNodeSync)) and rel.is_resource_pool:
                # If the relatiionship is a resource pool and the expected schema is different from the one of the pool, this means we expect to get
                # a resource from the pool itself
                allocate_from_pool = rel_schema.peer != rel.peer._schema.kind

            rel_data = rel._generate_input_data(allocate_from_pool=allocate_from_pool)

            if rel_data and isinstance(rel_data, dict):
                if variable_values := rel_data.get("data"):
                    data[item_name] = variable_values
                else:
                    data[item_name] = rel_data
                if variable_names := rel_data.get("variables"):
                    variables.update(variable_names)
            elif isinstance(rel_data, list):
                data[item_name] = rel_data
            elif rel_schema.cardinality == RelationshipCardinality.MANY:
                data[item_name] = []

        if exclude_unmodified:
            data, variables = self._strip_unmodified(data=data, variables=variables)

        mutation_variables = {key: type(value) for key, value in variables.items()}

        if self.id is not None:
            data["id"] = self.id
        elif self.hfid is not None and not exclude_hfid:
            data["hfid"] = self.hfid

        mutation_payload: dict[str, Any] = {"data": data}
        if context_data := self._get_request_context(request_context=request_context):
            mutation_payload["context"] = context_data

        # Add file variable for FileObject nodes with pending file content
        # file is a mutation argument at the same level as data, not inside data
        if self._file_object_support and self._file_content is not None:
            mutation_payload["file"] = "$file"
            mutation_variables["file"] = bytes

        return {
            "data": mutation_payload,
            "variables": variables,
            "mutation_variables": mutation_variables,
        }

    @staticmethod
    def _strip_unmodified_dict(data: dict, original_data: dict, variables: dict, item: str) -> None:
        data_item = data.get(item)
        if item in original_data and isinstance(original_data[item], dict) and isinstance(data_item, dict):
            for item_key in original_data[item]:
                for property_name in PROPERTIES_OBJECT:
                    if (
                        item_key == property_name
                        and isinstance(original_data[item][property_name], dict)
                        and original_data[item][property_name].get("id")
                    ):
                        original_data[item][property_name] = original_data[item][property_name]["id"]
                if item_key in data[item]:
                    if item_key == "id" and len(data[item].keys()) > 1:
                        # Related nodes typically require an ID. So the ID is only
                        # removed if it's the last key in the current context
                        continue

                    variable_key = None
                    if isinstance(data[item].get(item_key), str):
                        variable_key = data[item][item_key][1:]

                    if original_data[item].get(item_key) == data[item].get(item_key):
                        data[item].pop(item_key)
                    elif (
                        variable_key
                        and variable_key in variables
                        and original_data[item].get(item_key) == variables.get(variable_key)
                    ):
                        data[item].pop(item_key)
                        variables.pop(variable_key)

        # TODO: I do not feel _great_ about this
        # -> I don't even know who you are (but this is not great indeed) -- gmazoyer (quoting Thanos)
        original_data_item = original_data.get(item)
        original_data_item_is_none = original_data_item is None
        if isinstance(original_data_item, dict):
            if "node" in original_data_item:
                original_data_item_is_none = original_data_item["node"] is None
            elif "id" not in original_data_item:
                original_data_item_is_none = True

        if item in data and (data_item in ({}, []) or (data_item is None and original_data_item_is_none)):
            data.pop(item)

    def _strip_unmodified(self, data: dict, variables: dict) -> tuple[dict, dict]:
        original_data = self._data or {}
        for relationship in self._relationships:
            relationship_property = getattr(self, relationship)
            if not relationship_property or relationship not in data:
                continue
            if (
                not relationship_property.initialized
                and (
                    not isinstance(relationship_property, RelatedNodeBase) or not relationship_property.schema.optional
                )
            ) or (isinstance(relationship_property, RelationshipManagerBase) and not relationship_property.has_update):
                data.pop(relationship)

        for item in original_data:
            if item in data:
                if data[item] == original_data[item]:
                    if (
                        attr := getattr(self, item, None)
                    ) and (  # this should never be None, just a safety default value
                        not isinstance(attr, Attribute) or not attr.value_has_been_mutated
                    ):
                        data.pop(item)
                    continue
                if isinstance(original_data[item], dict):
                    self._strip_unmodified_dict(data=data, original_data=original_data, variables=variables, item=item)
                    if item in self._relationships and original_data[item].get("node"):
                        relationship_data_cardinality_one = copy(original_data)
                        relationship_data_cardinality_one[item] = original_data[item]["node"]
                        self._strip_unmodified_dict(
                            data=data, original_data=relationship_data_cardinality_one, variables=variables, item=item
                        )
                        # Run again to remove the "id" key if it's the last one remaining
                        self._strip_unmodified_dict(
                            data=data, original_data=relationship_data_cardinality_one, variables=variables, item=item
                        )

        return data, variables

    @staticmethod
    def _strip_alias(data: dict) -> dict[str, dict]:
        clean = {}

        under_node = False
        data_to_clean = data
        if "node" in data:
            under_node = True
            data_to_clean = data["node"]

        for key, value in data_to_clean.items():
            if "__alias__" in key:
                clean_key = key.split("__")[-1]
                clean[clean_key] = value
            else:
                clean[key] = value

        if under_node:
            complete = {k: v for k, v in data.items() if k != "node"}
            complete["node"] = clean
            return complete
        return clean

    def _validate_artifact_support(self, message: str) -> None:
        if not self._artifact_support:
            raise FeatureNotSupportedError(message)

    def _validate_hierarchy_support(self, message: str) -> None:
        if not self._hierarchy_support:
            raise FeatureNotSupportedError(message)

    def _validate_artifact_definition_support(self, message: str) -> None:
        if not self._artifact_definition_support:
            raise FeatureNotSupportedError(message)

    def _validate_file_object_support(self, message: str) -> None:
        if not self._file_object_support:
            raise FeatureNotSupportedError(message)

    def generate_query_data_init(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        partial_match: bool = False,
        order: Order | None = None,
        include_metadata: bool = False,
    ) -> dict[str, Any | dict]:
        data: dict[str, Any] = {
            "count": None,
            "edges": {"node": {"id": None, "hfid": None, "display_label": None, "__typename": None}},
        }

        if include_metadata:
            data["edges"]["node_metadata"] = NodeMetadata._generate_query_data()

        data["@filters"] = deepcopy(filters) if filters is not None else {}

        if order:
            data["@filters"]["order"] = order

        if offset is not None:
            data["@filters"]["offset"] = offset

        if limit is not None:
            data["@filters"]["limit"] = limit

        if include and exclude:
            in_both, _, _ = compare_lists(include, exclude)
            if in_both:
                raise ValueError(f"{in_both} are part of both include and exclude")

        if partial_match:
            data["@filters"]["partial_match"] = True

        return data

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (InfrahubNode, InfrahubNodeSync)):
            return NotImplemented
        return self.id == other.id

    def _relationship_mutation(self, action: str, relation_to_update: str, related_nodes: list[str]) -> str:
        related_node_str = ["{ id: " + f'"{node}"' + " }" for node in related_nodes]
        return f"""
        mutation {{
            Relationship{action}(
                data: {{
                    id: "{self.id}",
                    name: "{relation_to_update}",
                    nodes: [{", ".join(related_node_str)}]
                }}
            ) {{
                ok
            }}
        }}
        """

    def _get_attribute(self, name: str) -> Attribute:
        if name in self._attribute_data:
            return self._attribute_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have an attribute for {name}")


class InfrahubNode(InfrahubNodeBase):
    """Represents a Infrahub node in an asynchronous context."""

    def __init__(
        self,
        client: InfrahubClient,
        schema: MainSchemaTypesAPI,
        branch: str | None = None,
        data: dict | None = None,
    ) -> None:
        """
        Args:
            client: The client used to interact with the backend.
            schema: The schema of the node.
            branch: The branch where the node resides.
            data: Optional data to initialize the node.
        """
        self._client = client
        self._file_handler = FileHandler(client=client)

        # Extract node_metadata before extracting node data (node_metadata is sibling to node in edges)
        node_metadata_data: dict | None = None
        if isinstance(data, dict):
            node_metadata_data = data.get("node_metadata")
            if isinstance(data.get("node"), dict):
                data = data.get("node")

        self._relationship_cardinality_many_data: dict[str, RelationshipManager] = {}
        self._relationship_cardinality_one_data: dict[str, RelatedNode] = {}
        self._hierarchical_data: dict[str, RelatedNode | RelationshipManager] = {}

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

        # Initialize metadata after base class init
        if node_metadata_data:
            self._metadata = NodeMetadata(node_metadata_data)

    @classmethod
    async def from_graphql(
        cls,
        client: InfrahubClient,
        branch: str,
        data: dict,
        schema: MainSchemaTypesAPI | None = None,
        timeout: int | None = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename") or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = await client.schema.get(kind=node_kind, branch=branch, timeout=timeout)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: dict | RelatedNode | None = None) -> None:
        for rel_schema in self._schema.relationships:
            rel_data = data.get(rel_schema.name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                if isinstance(rel_data, RelatedNode):
                    peer_id_data: dict[str, Any] = {
                        key: value
                        for key, value in (
                            ("id", rel_data.id),
                            ("hfid", rel_data.hfid),
                            ("__typename", rel_data.typename),
                            ("kind", rel_data.kind),
                            ("display_label", rel_data.display_label),
                        )
                        if value is not None
                    }
                    rel_data = peer_id_data or None
                self._relationship_cardinality_one_data[rel_schema.name] = RelatedNode(
                    name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=rel_data
                )
            else:
                self._relationship_cardinality_many_data[rel_schema.name] = RelationshipManager(
                    name=rel_schema.name,
                    client=self._client,
                    node=self,
                    branch=self._branch,
                    schema=rel_schema,
                    data=rel_data,
                )
        # Initialize parent, children, ancestors and descendants for hierarchical nodes
        if self._hierarchy_support:
            # Create pseudo-schema for parent (cardinality one)
            parent_schema = RelationshipSchemaAPI(
                name="parent",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                kind=RelationshipKind.HIERARCHY,
                cardinality="one",
                optional=True,
            )
            parent_data = data.get("parent", None) if isinstance(data, dict) else None
            self._hierarchical_data["parent"] = RelatedNode(
                name="parent",
                client=self._client,
                branch=self._branch,
                schema=parent_schema,
                data=parent_data,
            )
            # Create pseudo-schema for children (many cardinality)
            children_schema = RelationshipSchemaAPI(
                name="children",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                kind=RelationshipKind.HIERARCHY,
                cardinality="many",
                optional=True,
            )
            children_data = data.get("children", None) if isinstance(data, dict) else None
            self._hierarchical_data["children"] = RelationshipManager(
                name="children",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=children_schema,
                data=children_data,
            )
            # Create pseudo-schema for ancestors (read-only, many cardinality)
            ancestors_schema = RelationshipSchemaAPI(
                name="ancestors",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                cardinality="many",
                read_only=True,
                optional=True,
            )
            ancestors_data = data.get("ancestors", None) if isinstance(data, dict) else None
            self._hierarchical_data["ancestors"] = RelationshipManager(
                name="ancestors",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=ancestors_schema,
                data=ancestors_data,
            )
            # Create pseudo-schema for descendants (read-only, many cardinality)
            descendants_schema = RelationshipSchemaAPI(
                name="descendants",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                cardinality="many",
                read_only=True,
                optional=True,
            )
            descendants_data = data.get("descendants", None) if isinstance(data, dict) else None
            self._hierarchical_data["descendants"] = RelationshipManager(
                name="descendants",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=descendants_schema,
                data=descendants_data,
            )

    def __getattr__(self, name: str) -> Attribute | RelationshipManager | RelatedNode:
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            return self._attribute_data[name]
        if "_relationship_cardinality_many_data" in self.__dict__ and name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]
        if "_hierarchical_data" in self.__dict__ and name in self._hierarchical_data:
            return self._hierarchical_data[name]

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for relationship names that exist or revert to normal behaviour"""
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            rel_schemas = [rel_schema for rel_schema in self._schema.relationships if rel_schema.name == name]
            if not rel_schemas:
                raise SchemaNotFoundError(
                    identifier=self._schema.kind,
                    message=f"Unable to find relationship schema for '{name}' on {self._schema.kind}",
                )
            rel_schema = rel_schemas[0]
            self._relationship_cardinality_one_data[name] = RelatedNode(
                name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=value
            )
            return

        super().__setattr__(name, value)

    async def generate(self, nodes: list[str] | None = None) -> None:
        self._validate_artifact_definition_support(ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        nodes = nodes or []
        payload = {"nodes": nodes}
        resp = await self._client._post(f"{self._client.address}/api/artifact/generate/{self.id}", payload=payload)
        resp.raise_for_status()

    async def artifact_generate(self, name: str) -> None:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        artifact = await self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        await artifact._get_relationship_one(name="definition").fetch()
        await artifact._get_relationship_one(name="definition").peer.generate([artifact.id])

    async def artifact_fetch(self, name: str) -> str | dict[str, Any]:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        artifact = await self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        return await self._client.object_store.get(identifier=artifact._get_attribute(name="storage_id").value)

    async def download_file(self, dest: Path | None = None) -> bytes | int:
        """Download the file content from this FileObject node.

        This method is only available for nodes that inherit from CoreFileObject.
        The node must have been saved (have an id) before calling this method.

        Args:
            dest: Optional destination path. If provided, the file will be streamed
                  directly to this path (memory-efficient for large files) and the
                  number of bytes written will be returned. If not provided, the
                  file content will be returned as bytes.

        Returns:
            If ``dest`` is None: The file content as bytes.
            If ``dest`` is provided: The number of bytes written to the file.

        Raises:
            FeatureNotSupportedError: If this node doesn't inherit from CoreFileObject.
            ValueError: If the node hasn't been saved yet or file not found.
            AuthenticationError: If authentication fails.

        Examples:
            >>> # Download to memory
            >>> content = await contract.download_file()

            >>> # Stream to file (memory-efficient for large files)
            >>> bytes_written = await contract.download_file(dest=Path("/tmp/contract.pdf"))
        """
        self._validate_file_object_support(message=FILE_DOWNLOAD_FEATURE_NOT_SUPPORTED_MESSAGE)

        if not self.id:
            raise ValueError("Cannot download file for a node that hasn't been saved yet.")

        return await self._file_handler.download(node_id=self.id, branch=self._branch, dest=dest)

    async def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None:
        input_data = {"data": {"id": self.id}}
        if context_data := self._get_request_context(request_context=request_context):
            input_data["context"] = context_data

        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            timeout=timeout,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
        )

    async def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None:
        if self._existing is False or allow_upsert is True:
            await self.create(allow_upsert=allow_upsert, timeout=timeout, request_context=request_context)
        else:
            await self.update(timeout=timeout, request_context=request_context)

        if update_group_context is None and self._client.mode == InfrahubClientMode.TRACKING:
            update_group_context = True

        if not isinstance(self._schema, GenericSchemaAPI):
            if "CoreGroup" in self._schema.inherit_from:
                await self._client.group_context.add_related_groups(
                    ids=[self.id], update_group_context=update_group_context
                )
            else:
                await self._client.group_context.add_related_nodes(
                    ids=[self.id], update_group_context=update_group_context
                )
        else:
            await self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)

        self._client.store.set(node=self)

    async def _process_hierarchical_fields(
        self,
        data: dict[str, Any],
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        prefetch_relationships: bool = False,
        insert_alias: bool = False,
        property: bool = False,
    ) -> None:
        """Process hierarchical fields (parent, children, ancestors, descendants) for hierarchical nodes."""
        if not self._hierarchy_support:
            return

        for hierarchical_name in ["parent", "children", "ancestors", "descendants"]:
            if exclude and hierarchical_name in exclude:
                continue

            # Only include if explicitly requested or if prefetch_relationships is True
            should_fetch = prefetch_relationships or (include is not None and hierarchical_name in include)
            if not should_fetch:
                continue

            peer_schema = await self._client.schema.get(kind=self._schema.hierarchy, branch=self._branch)  # type: ignore[union-attr, arg-type]
            peer_node = InfrahubNode(client=self._client, schema=peer_schema, branch=self._branch)
            # Exclude hierarchical fields from peer data to prevent infinite recursion
            peer_exclude = list(exclude) if exclude else []
            peer_exclude.extend(["parent", "children", "ancestors", "descendants"])
            peer_data = await peer_node.generate_query_data_node(
                exclude=peer_exclude,
                property=property,
            )

            # Parent is cardinality one, others are cardinality many
            if hierarchical_name == "parent":
                hierarchical_data = RelatedNode._generate_query_data(peer_data=peer_data, property=property)
                # Use fragment for hierarchical fields similar to hierarchy relationships
                data_node = hierarchical_data["node"]
                hierarchical_data["node"] = {}
                hierarchical_data["node"][f"...on {self._schema.hierarchy}"] = data_node  # type: ignore[union-attr]
            else:
                hierarchical_data = RelationshipManager._generate_query_data(peer_data=peer_data, property=property)
                # Use fragment for hierarchical fields similar to hierarchy relationships
                data_node = hierarchical_data["edges"]["node"]
                hierarchical_data["edges"]["node"] = {}
                hierarchical_data["edges"]["node"][f"...on {self._schema.hierarchy}"] = data_node  # type: ignore[union-attr]

            data[hierarchical_name] = hierarchical_data

            if insert_alias:
                data[hierarchical_name]["@alias"] = f"__alias__{self._schema.kind}__{hierarchical_name}"

    async def generate_query_data(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        fragment: bool = False,
        prefetch_relationships: bool = False,
        partial_match: bool = False,
        property: bool = False,
        order: Order | None = None,
        include_metadata: bool = False,
    ) -> dict[str, Any | dict]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            partial_match=partial_match,
            order=order,
            include_metadata=include_metadata,
        )
        data["edges"]["node"].update(
            await self.generate_query_data_node(
                include=include,
                exclude=exclude,
                prefetch_relationships=prefetch_relationships,
                inherited=True,
                property=property,
                include_metadata=include_metadata,
            )
        )

        if isinstance(self._schema, GenericSchemaAPI) and fragment:
            for child in self._schema.used_by:
                child_schema = await self._client.schema.get(kind=child)
                child_node = InfrahubNode(client=self._client, schema=child_schema)

                # Add the attribute and the relationship already part of the parent to the exclude list for the children
                exclude_parent = self._attributes + self._relationships
                _, _, only_in_list2 = compare_lists(list1=include or [], list2=exclude_parent)

                exclude_child = only_in_list2
                if exclude:
                    exclude_child += exclude

                child_data = await child_node.generate_query_data_node(
                    include=include,
                    exclude=exclude_child,
                    prefetch_relationships=prefetch_relationships,
                    inherited=False,
                    insert_alias=True,
                    property=property,
                    include_metadata=include_metadata,
                )

                if child_data:
                    data["edges"]["node"][f"...on {child}"] = child_data

        return {self._schema.kind: data}

    async def generate_query_data_node(
        self,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        inherited: bool = True,
        insert_alias: bool = False,
        prefetch_relationships: bool = False,
        property: bool = False,
        include_metadata: bool = False,
    ) -> dict[str, Any | dict]:
        """Generate the node part of a GraphQL Query with attributes and nodes.

        Args:
            include (Optional[list[str]], optional): List of attributes or relationships to include. Defaults to None.
            exclude (Optional[list[str]], optional): List of attributes or relationships to exclude. Defaults to None.
            inherited (bool, optional): Indicated of the attributes and the relationships inherited from generics should be included as well.
                                        Defaults to True.
            insert_alias (bool, optional): If True, inserts aliases in the query for each attribute or relationship.
            prefetch_relationships (bool, optional): If True, pre-fetches relationship data as part of the query.
            include_metadata (bool, optional): If True, includes node_metadata and relationship_metadata in the query.

        Returns:
            dict[str, Union[Any, Dict]]: GraphQL query in dictionary format
        """

        data: dict[str, Any] = {}

        for attr_name in self._attributes:
            if exclude and attr_name in exclude:
                continue

            attr: Attribute = getattr(self, attr_name)

            if not inherited and attr._schema.inherited:
                continue

            attr_data = attr._generate_query_data(property=property, include_metadata=include_metadata)
            if attr_data:
                data[attr_name] = attr_data
                if insert_alias:
                    data[attr_name]["@alias"] = f"__alias__{self._schema.kind}__{attr_name}"
            elif insert_alias:
                if insert_alias:
                    data[attr_name] = {"@alias": f"__alias__{self._schema.kind}__{attr_name}"}

        for rel_name in self._relationships:
            if exclude and rel_name in exclude:
                continue

            rel_schema = self._schema.get_relationship(name=rel_name)

            if not rel_schema or (not inherited and rel_schema.inherited):
                continue

            if (
                rel_schema.cardinality == RelationshipCardinality.MANY  # type: ignore[union-attr]
                and rel_schema.kind not in {RelationshipKind.ATTRIBUTE, RelationshipKind.PARENT}  # type: ignore[union-attr]
                and not (include and rel_name in include)
            ):
                continue

            peer_data: dict[str, Any] = {}
            should_fetch_relationship = prefetch_relationships or (include is not None and rel_name in include)
            if rel_schema and should_fetch_relationship:
                peer_schema = await self._client.schema.get(kind=rel_schema.peer, branch=self._branch)
                peer_node = InfrahubNode(client=self._client, schema=peer_schema, branch=self._branch)
                peer_data = await peer_node.generate_query_data_node(
                    property=property,
                    include_metadata=include_metadata,
                )

            rel_data: dict[str, Any]
            if rel_schema and rel_schema.cardinality == "one":
                rel_data = RelatedNode._generate_query_data(
                    peer_data=peer_data, property=property, include_metadata=include_metadata
                )
                # Nodes involved in a hierarchy are required to inherit from a common ancestor node, and graphql
                # tries to resolve attributes in this ancestor instead of actual node. To avoid
                # invalid queries issues when attribute is missing in the common ancestor, we use a fragment
                # to explicit actual node kind we are querying.
                if rel_schema.kind == RelationshipKind.HIERARCHY:
                    data_node = rel_data["node"]
                    rel_data["node"] = {}
                    rel_data["node"][f"...on {rel_schema.peer}"] = data_node
            elif rel_schema and rel_schema.cardinality == "many":
                rel_data = RelationshipManager._generate_query_data(
                    peer_data=peer_data, property=property, include_metadata=include_metadata
                )
            else:
                continue

            data[rel_name] = rel_data

            if insert_alias:
                data[rel_name]["@alias"] = f"__alias__{self._schema.kind}__{rel_name}"

        # Add parent, children, ancestors and descendants for hierarchical nodes
        await self._process_hierarchical_fields(
            data=data,
            include=include,
            exclude=exclude,
            prefetch_relationships=prefetch_relationships,
            insert_alias=insert_alias,
            property=property,
        )

        return data

    async def add_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Add", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipadd-{relation_to_update}"
        await self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    async def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Remove", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipremove-{relation_to_update}"
        await self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def _generate_mutation_query(self) -> dict[str, Any]:
        query_result: dict[str, Any] = {"ok": None, "object": {"id": None}}

        for attr_name in self._attributes:
            attr: Attribute = getattr(self, attr_name)
            query_result["object"].update(attr._generate_mutation_query())

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if not isinstance(rel, RelatedNode):
                continue

            query_result["object"].update(rel._generate_mutation_query())

        return query_result

    async def _process_mutation_result(
        self, mutation_name: str, response: dict[str, Any], timeout: int | None = None
    ) -> None:
        object_response: dict[str, Any] = response[mutation_name]["object"]
        self.id = object_response["id"]
        self._existing = True

        for attr_name in self._attributes:
            attr = getattr(self, attr_name)
            if attr_name not in object_response or not attr.is_from_pool_attribute():
                continue

            # Process allocated resource from a pool and update attribute
            attr.value = object_response[attr_name]["value"]

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel_name not in object_response or not isinstance(rel, RelatedNode) or not rel.is_resource_pool:
                continue

            # Process allocated resource from a pool and update related node
            allocated_resource = object_response[rel_name]
            related_node = RelatedNode(
                client=self._client, branch=self._branch, schema=rel.schema, data=allocated_resource
            )
            await related_node.fetch(timeout=timeout)
            setattr(self, rel_name, related_node)

    async def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        if self._file_object_support and self._file_content is None:
            raise ValueError(
                f"Cannot create {self._schema.kind} without file content. Use upload_from_path() or upload_from_bytes() to provide "
                "file content before saving."
            )

        mutation_query = self._generate_mutation_query()

        # Upserting means we may want to create, meaning payload contains all mandatory fields required for a creation,
        # so hfid is just redondant information. Currently, upsert mutation has performance overhead if `hfid` is filled.
        if allow_upsert:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Upsert"
            tracker = f"mutation-{str(self._schema.kind).lower()}-upsert"
        else:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Create"
            tracker = f"mutation-{str(self._schema.kind).lower()}-create"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        if "file" in input_data["mutation_variables"]:
            prepared = await self._get_file_for_upload()
            try:
                response = await self._client._execute_graphql_with_file(
                    query=query.render(),
                    variables=input_data["variables"],
                    file_content=prepared.file_object,
                    file_name=prepared.filename,
                    branch_name=self._branch,
                    tracker=tracker,
                    timeout=timeout,
                )
            finally:
                if prepared.should_close and prepared.file_object:
                    prepared.file_object.close()
            # Clear the file content after successful upload
            self.clear_file()
        else:
            response = await self._client.execute_graphql(
                query=query.render(),
                branch_name=self._branch,
                tracker=tracker,
                variables=input_data["variables"],
                timeout=timeout,
            )
        await self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    async def update(
        self, do_full_update: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        input_data = self._generate_input_data(exclude_unmodified=not do_full_update, request_context=request_context)
        mutation_query = self._generate_mutation_query()
        mutation_name = f"{self._schema.kind}Update"
        tracker = f"mutation-{str(self._schema.kind).lower()}-update"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        if "file" in input_data["mutation_variables"]:
            prepared = await self._get_file_for_upload()
            try:
                response = await self._client._execute_graphql_with_file(
                    query=query.render(),
                    variables=input_data["variables"],
                    file_content=prepared.file_object,
                    file_name=prepared.filename,
                    branch_name=self._branch,
                    tracker=tracker,
                    timeout=timeout,
                )
            finally:
                if prepared.should_close and prepared.file_object:
                    prepared.file_object.close()
            # Clear the file content after successful upload
            self.clear_file()
        else:
            response = await self._client.execute_graphql(
                query=query.render(),
                branch_name=self._branch,
                timeout=timeout,
                tracker=tracker,
                variables=input_data["variables"],
            )
        await self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    async def _process_relationships(
        self,
        node_data: dict[str, Any],
        branch: str,
        related_nodes: list[InfrahubNode],
        timeout: int | None = None,
        recursive: bool = False,
    ) -> None:
        """Processes the Relationships of a InfrahubNode and add Related Nodes to a list.

        Args:
            node_data (dict[str, Any]): The item from the GraphQL response corresponding to the node.
            branch (str): The branch name.
            related_nodes (list[InfrahubNode]): The list to which related nodes will be appended.
            timeout (int, optional): Overrides default timeout used when querying the graphql API. Specified in seconds.
            recursive:(bool): Whether to recursively process relationships of related nodes.
        """
        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel and isinstance(rel, RelatedNode):
                relation = node_data["node"].get(rel_name, None)
                if relation and relation.get("node", None):
                    related_node = await InfrahubNode.from_graphql(
                        client=self._client,
                        branch=branch,
                        data=relation,
                        timeout=timeout,
                    )
                    related_nodes.append(related_node)
                    if recursive:
                        await related_node._process_relationships(
                            node_data=relation,
                            branch=branch,
                            related_nodes=related_nodes,
                            recursive=recursive,
                        )
            elif rel and isinstance(rel, RelationshipManager):
                peers = node_data["node"].get(rel_name, None)
                if peers and peers["edges"]:
                    for peer in peers["edges"]:
                        related_node = await InfrahubNode.from_graphql(
                            client=self._client,
                            branch=branch,
                            data=peer,
                            timeout=timeout,
                        )
                        related_nodes.append(related_node)
                        if recursive:
                            await related_node._process_relationships(
                                node_data=peer,
                                branch=branch,
                                related_nodes=related_nodes,
                                recursive=recursive,
                            )

    async def get_pool_allocated_resources(self, resource: InfrahubNode) -> list[InfrahubNode]:
        """Fetch all nodes that were allocated for the pool and a given resource.

        Args:
            resource (InfrahubNode): The resource from which the nodes were allocated.

        Returns:
            list[InfrahubNode]: The allocated nodes.
        """
        if not self.is_resource_pool():
            raise ValueError("Allocated resources can only be fetched from resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolAllocated"
        node_ids_per_kind: dict[str, list[str]] = {}

        has_remaining_items = True
        page_number = 1
        while has_remaining_items:
            page_offset = (page_number - 1) * self._client.pagination_size

            query = Query(
                query={
                    graphql_query_name: {
                        "@filters": {
                            "pool_id": "$pool_id",
                            "resource_id": "$resource_id",
                            "offset": page_offset,
                            "limit": self._client.pagination_size,
                        },
                        "count": None,
                        "edges": {"node": {"id": None, "kind": None, "branch": None, "identifier": None}},
                    }
                },
                name="GetAllocatedResourceForPool",
                variables={"pool_id": str, "resource_id": str},
            )
            response = await self._client.execute_graphql(
                query=query.render(),
                variables={"pool_id": self.id, "resource_id": resource.id},
                branch_name=self._branch,
                tracker=f"get-allocated-resources-page{page_number}",
            )

            for edge in response[graphql_query_name]["edges"]:
                node = edge["node"]
                node_ids_per_kind.setdefault(node["kind"], []).append(node["id"])

            remaining_items = response[graphql_query_name].get("count", 0) - (
                page_offset + self._client.pagination_size
            )
            if remaining_items <= 0:
                has_remaining_items = False

            page_number += 1

        nodes: list[InfrahubNode] = []
        for kind, node_ids in node_ids_per_kind.items():
            nodes.extend(await self._client.filters(kind=kind, branch=self._branch, ids=node_ids))

        return nodes

    async def get_pool_resources_utilization(self) -> list[dict[str, Any]]:
        """Fetch the utilization of each resource for the pool.

        Returns:
            list[dict[str, Any]]: A list containing the allocation numbers for each resource of the pool.
        """
        if not self.is_resource_pool():
            raise ValueError("Pool utilization can only be fetched for resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolUtilization"

        query = Query(
            query={
                graphql_query_name: {
                    "@filters": {"pool_id": "$pool_id"},
                    "count": None,
                    "edges": {
                        "node": {
                            "id": None,
                            "kind": None,
                            "utilization": None,
                            "utilization_branches": None,
                            "utilization_default_branch": None,
                        }
                    },
                }
            },
            name="GetUtilizationForPool",
            variables={"pool_id": str},
        )
        response = await self._client.execute_graphql(
            query=query.render(),
            variables={"pool_id": self.id},
            branch_name=self._branch,
            tracker="get-pool-utilization",
        )

        if response[graphql_query_name].get("count", 0):
            return [edge["node"] for edge in response[graphql_query_name]["edges"]]
        return []

    def _get_relationship_many(self, name: str) -> RelationshipManager:
        if name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=many relationship for {name}")

    def _get_relationship_one(self, name: str) -> RelatedNode:
        if name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=one relationship for {name}")

    async def get_flat_value(self, key: str, separator: str = "__") -> Any:
        """Query recursively a value defined in a flat notation (string), on a hierarchy of objects

        Examples:
            name__value
            module.object.value
        """
        if separator not in key:
            return getattr(self, key)

        first, remaining = key.split(separator, maxsplit=1)

        if first in self._schema.attribute_names:
            attr = getattr(self, first)
            for part in remaining.split(separator):
                attr = getattr(attr, part)
            return attr

        try:
            rel = self._schema.get_relationship(name=first)
        except ValueError as exc:
            raise ValueError(f"No attribute or relationship named '{first}' for '{self._schema.kind}'") from exc

        if rel.cardinality != RelationshipCardinality.ONE:
            raise ValueError(
                f"Can only look up flat value for relationships of cardinality {RelationshipCardinality.ONE.value}"
            )

        related_node: RelatedNode = getattr(self, first)
        await related_node.fetch()
        return await related_node.peer.get_flat_value(key=remaining, separator=separator)

    async def extract(self, params: dict[str, str]) -> dict[str, Any]:
        """Extract some data points defined in a flat notation."""
        result: dict[str, Any] = {}
        for key, value in params.items():
            result[key] = await self.get_flat_value(key=value)

        return result

    def __dir__(self) -> Iterable[str]:
        base = list(super().__dir__())
        return sorted(
            base
            + list(self._attribute_data.keys())
            + list(self._relationship_cardinality_many_data.keys())
            + list(self._relationship_cardinality_one_data.keys())
        )


class InfrahubNodeSync(InfrahubNodeBase):
    """Represents a Infrahub node in a synchronous context."""

    def __init__(
        self,
        client: InfrahubClientSync,
        schema: MainSchemaTypesAPI,
        branch: str | None = None,
        data: dict | None = None,
    ) -> None:
        """
        Args:
            client (InfrahubClientSync): The client used to interact with the backend synchronously.
            schema (MainSchemaTypes): The schema of the node.
            branch (Optional[str]): The branch where the node resides.
            data (Optional[dict]): Optional data to initialize the node.
        """
        self._client = client
        self._file_handler = FileHandlerSync(client=client)

        # Extract node_metadata before extracting node data (node_metadata is sibling to node in edges)
        node_metadata_data: dict | None = None
        if isinstance(data, dict):
            node_metadata_data = data.get("node_metadata")
            if isinstance(data.get("node"), dict):
                data = data.get("node")

        self._relationship_cardinality_many_data: dict[str, RelationshipManagerSync] = {}
        self._relationship_cardinality_one_data: dict[str, RelatedNodeSync] = {}
        self._hierarchical_data: dict[str, RelatedNodeSync | RelationshipManagerSync] = {}

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

        # Initialize metadata after base class init
        if node_metadata_data:
            self._metadata = NodeMetadata(node_metadata_data)

    @classmethod
    def from_graphql(
        cls,
        client: InfrahubClientSync,
        branch: str,
        data: dict,
        schema: MainSchemaTypesAPI | None = None,
        timeout: int | None = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename") or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = client.schema.get(kind=node_kind, branch=branch, timeout=timeout)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: dict | None = None) -> None:
        for rel_schema in self._schema.relationships:
            rel_data = data.get(rel_schema.name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                if isinstance(rel_data, RelatedNodeSync):
                    peer_id_data: dict[str, Any] = {
                        key: value
                        for key, value in (
                            ("id", rel_data.id),
                            ("hfid", rel_data.hfid),
                            ("__typename", rel_data.typename),
                            ("kind", rel_data.kind),
                            ("display_label", rel_data.display_label),
                        )
                        if value is not None
                    }
                    rel_data = peer_id_data or None
                self._relationship_cardinality_one_data[rel_schema.name] = RelatedNodeSync(
                    name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=rel_data
                )
            else:
                self._relationship_cardinality_many_data[rel_schema.name] = RelationshipManagerSync(
                    name=rel_schema.name,
                    client=self._client,
                    node=self,
                    branch=self._branch,
                    schema=rel_schema,
                    data=rel_data,
                )

        # Initialize parent, children, ancestors and descendants for hierarchical nodes
        if self._hierarchy_support:
            # Create pseudo-schema for parent (cardinality one)
            parent_schema = RelationshipSchemaAPI(
                name="parent",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                kind=RelationshipKind.HIERARCHY,
                cardinality="one",
                optional=True,
            )
            parent_data = data.get("parent", None) if isinstance(data, dict) else None
            self._hierarchical_data["parent"] = RelatedNodeSync(
                name="parent",
                client=self._client,
                branch=self._branch,
                schema=parent_schema,
                data=parent_data,
            )

            # Create pseudo-schema for children (many cardinality)
            children_schema = RelationshipSchemaAPI(
                name="children",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                kind=RelationshipKind.HIERARCHY,
                cardinality="many",
                optional=True,
            )
            children_data = data.get("children", None) if isinstance(data, dict) else None
            self._hierarchical_data["children"] = RelationshipManagerSync(
                name="children",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=children_schema,
                data=children_data,
            )

            # Create pseudo-schema for ancestors (read-only, many cardinality)
            ancestors_schema = RelationshipSchemaAPI(
                name="ancestors",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                cardinality="many",
                read_only=True,
                optional=True,
            )
            ancestors_data = data.get("ancestors", None) if isinstance(data, dict) else None
            self._hierarchical_data["ancestors"] = RelationshipManagerSync(
                name="ancestors",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=ancestors_schema,
                data=ancestors_data,
            )

            # Create pseudo-schema for descendants (read-only, many cardinality)
            descendants_schema = RelationshipSchemaAPI(
                name="descendants",
                peer=self._schema.hierarchy,  # type: ignore[union-attr, arg-type]
                cardinality="many",
                read_only=True,
                optional=True,
            )
            descendants_data = data.get("descendants", None) if isinstance(data, dict) else None
            self._hierarchical_data["descendants"] = RelationshipManagerSync(
                name="descendants",
                client=self._client,
                node=self,
                branch=self._branch,
                schema=descendants_schema,
                data=descendants_data,
            )

    def __getattr__(self, name: str) -> Attribute | RelationshipManagerSync | RelatedNodeSync:
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            return self._attribute_data[name]
        if "_relationship_cardinality_many_data" in self.__dict__ and name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]
        if "_hierarchical_data" in self.__dict__ and name in self._hierarchical_data:
            return self._hierarchical_data[name]

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for relationship names that exist or revert to normal behaviour"""
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            rel_schemas = [rel_schema for rel_schema in self._schema.relationships if rel_schema.name == name]
            if not rel_schemas:
                raise SchemaNotFoundError(
                    identifier=self._schema.kind,
                    message=f"Unable to find relationship schema for '{name}' on {self._schema.kind}",
                )
            rel_schema = rel_schemas[0]
            self._relationship_cardinality_one_data[name] = RelatedNodeSync(
                name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=value
            )
            return

        super().__setattr__(name, value)

    def generate(self, nodes: list[str] | None = None) -> None:
        self._validate_artifact_definition_support(ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)
        nodes = nodes or []
        payload = {"nodes": nodes}
        resp = self._client._post(f"{self._client.address}/api/artifact/generate/{self.id}", payload=payload)
        resp.raise_for_status()

    def artifact_generate(self, name: str) -> None:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)
        artifact = self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        artifact._get_relationship_one(name="definition").fetch()
        artifact._get_relationship_one(name="definition").peer.generate([artifact.id])

    def artifact_fetch(self, name: str) -> str | dict[str, Any]:
        self._validate_artifact_support(ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE)
        artifact = self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        return self._client.object_store.get(identifier=artifact._get_attribute(name="storage_id").value)

    def download_file(self, dest: Path | None = None) -> bytes | int:
        """Download the file content from this FileObject node.

        This method is only available for nodes that inherit from CoreFileObject.
        The node must have been saved (have an id) before calling this method.

        Args:
            dest: Optional destination path. If provided, the file will be streamed
                  directly to this path (memory-efficient for large files) and the
                  number of bytes written will be returned. If not provided, the
                  file content will be returned as bytes.

        Returns:
            If ``dest`` is None: The file content as bytes.
            If ``dest`` is provided: The number of bytes written to the file.

        Raises:
            FeatureNotSupportedError: If this node doesn't inherit from CoreFileObject.
            ValueError: If the node hasn't been saved yet or file not found.
            AuthenticationError: If authentication fails.

        Examples:
            >>> # Download to memory
            >>> content = contract.download_file()

            >>> # Stream to file (memory-efficient for large files)
            >>> bytes_written = contract.download_file(dest=Path("/tmp/contract.pdf"))
        """
        self._validate_file_object_support(message=FILE_DOWNLOAD_FEATURE_NOT_SUPPORTED_MESSAGE)

        if not self.id:
            raise ValueError("Cannot download file for a node that hasn't been saved yet.")

        return self._file_handler.download(node_id=self.id, branch=self._branch, dest=dest)

    def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None:
        input_data = {"data": {"id": self.id}}
        if context_data := self._get_request_context(request_context=request_context):
            input_data["context"] = context_data

        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
            timeout=timeout,
        )

    def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None:
        if self._existing is False or allow_upsert is True:
            self.create(allow_upsert=allow_upsert, timeout=timeout, request_context=request_context)
        else:
            self.update(timeout=timeout, request_context=request_context)

        if update_group_context is None and self._client.mode == InfrahubClientMode.TRACKING:
            update_group_context = True

        if not isinstance(self._schema, GenericSchemaAPI):
            if "CoreGroup" in self._schema.inherit_from:
                self._client.group_context.add_related_groups(ids=[self.id], update_group_context=update_group_context)
            else:
                self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)
        else:
            self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)

        self._client.store.set(node=self)

    def _process_hierarchical_fields(
        self,
        data: dict[str, Any],
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        prefetch_relationships: bool = False,
        insert_alias: bool = False,
        property: bool = False,
    ) -> None:
        """Process hierarchical fields (parent, children, ancestors, descendants) for hierarchical nodes."""
        if not self._hierarchy_support:
            return

        for hierarchical_name in ["parent", "children", "ancestors", "descendants"]:
            if exclude and hierarchical_name in exclude:
                continue

            # Only include if explicitly requested or if prefetch_relationships is True
            should_fetch = prefetch_relationships or (include is not None and hierarchical_name in include)
            if not should_fetch:
                continue

            peer_schema = self._client.schema.get(kind=self._schema.hierarchy, branch=self._branch)  # type: ignore[union-attr, arg-type]
            peer_node = InfrahubNodeSync(client=self._client, schema=peer_schema, branch=self._branch)
            # Exclude hierarchical fields from peer data to prevent infinite recursion
            peer_exclude = list(exclude) if exclude else []
            peer_exclude.extend(["parent", "children", "ancestors", "descendants"])
            peer_data = peer_node.generate_query_data_node(
                exclude=peer_exclude,
                property=property,
            )

            # Parent is cardinality one, others are cardinality many
            if hierarchical_name == "parent":
                hierarchical_data = RelatedNodeSync._generate_query_data(peer_data=peer_data, property=property)
                # Use fragment for hierarchical fields similar to hierarchy relationships
                data_node = hierarchical_data["node"]
                hierarchical_data["node"] = {}
                hierarchical_data["node"][f"...on {self._schema.hierarchy}"] = data_node  # type: ignore[union-attr]
            else:
                hierarchical_data = RelationshipManagerSync._generate_query_data(peer_data=peer_data, property=property)
                # Use fragment for hierarchical fields similar to hierarchy relationships
                data_node = hierarchical_data["edges"]["node"]
                hierarchical_data["edges"]["node"] = {}
                hierarchical_data["edges"]["node"][f"...on {self._schema.hierarchy}"] = data_node  # type: ignore[union-attr]

            data[hierarchical_name] = hierarchical_data

            if insert_alias:
                data[hierarchical_name]["@alias"] = f"__alias__{self._schema.kind}__{hierarchical_name}"

    def generate_query_data(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        fragment: bool = False,
        prefetch_relationships: bool = False,
        partial_match: bool = False,
        property: bool = False,
        order: Order | None = None,
        include_metadata: bool = False,
    ) -> dict[str, Any | dict]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            partial_match=partial_match,
            order=order,
            include_metadata=include_metadata,
        )
        data["edges"]["node"].update(
            self.generate_query_data_node(
                include=include,
                exclude=exclude,
                prefetch_relationships=prefetch_relationships,
                inherited=True,
                property=property,
                include_metadata=include_metadata,
            )
        )

        if isinstance(self._schema, GenericSchemaAPI) and fragment:
            for child in self._schema.used_by:
                child_schema = self._client.schema.get(kind=child)
                child_node = InfrahubNodeSync(client=self._client, schema=child_schema)

                exclude_parent = self._attributes + self._relationships
                _, _, only_in_list2 = compare_lists(list1=include or [], list2=exclude_parent)

                exclude_child = only_in_list2
                if exclude:
                    exclude_child += exclude

                child_data = child_node.generate_query_data_node(
                    include=include,
                    exclude=exclude_child,
                    prefetch_relationships=prefetch_relationships,
                    inherited=False,
                    insert_alias=True,
                    property=property,
                    include_metadata=include_metadata,
                )

                if child_data:
                    data["edges"]["node"][f"...on {child}"] = child_data

        return {self._schema.kind: data}

    def generate_query_data_node(
        self,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        inherited: bool = True,
        insert_alias: bool = False,
        prefetch_relationships: bool = False,
        property: bool = False,
        include_metadata: bool = False,
    ) -> dict[str, Any | dict]:
        """Generate the node part of a GraphQL Query with attributes and nodes.

        Args:
            include (Optional[list[str]], optional): List of attributes or relationships to include. Defaults to None.
            exclude (Optional[list[str]], optional): List of attributes or relationships to exclude. Defaults to None.
            inherited (bool, optional): Indicated of the attributes and the relationships inherited from generics should be included as well.
                                        Defaults to True.
            insert_alias (bool, optional): If True, inserts aliases in the query for each attribute or relationship.
            prefetch_relationships (bool, optional): If True, pre-fetches relationship data as part of the query.
            include_metadata (bool, optional): If True, includes node_metadata and relationship_metadata in the query.

        Returns:
            dict[str, Union[Any, Dict]]: GraphQL query in dictionary format
        """

        data: dict[str, Any] = {}

        for attr_name in self._attributes:
            if exclude and attr_name in exclude:
                continue

            attr: Attribute = getattr(self, attr_name)

            if not inherited and attr._schema.inherited:
                continue

            attr_data = attr._generate_query_data(property=property, include_metadata=include_metadata)
            if attr_data:
                data[attr_name] = attr_data
                if insert_alias:
                    data[attr_name]["@alias"] = f"__alias__{self._schema.kind}__{attr_name}"
            elif insert_alias:
                if insert_alias:
                    data[attr_name] = {"@alias": f"__alias__{self._schema.kind}__{attr_name}"}

        for rel_name in self._relationships:
            if exclude and rel_name in exclude:
                continue

            rel_schema = self._schema.get_relationship(name=rel_name)

            if not rel_schema or (not inherited and rel_schema.inherited):
                continue

            if (
                rel_schema.cardinality == RelationshipCardinality.MANY  # type: ignore[union-attr]
                and rel_schema.kind not in {RelationshipKind.ATTRIBUTE, RelationshipKind.PARENT}  # type: ignore[union-attr]
                and not (include and rel_name in include)
            ):
                continue

            peer_data: dict[str, Any] = {}
            should_fetch_relationship = prefetch_relationships or (include is not None and rel_name in include)
            if rel_schema and should_fetch_relationship:
                peer_schema = self._client.schema.get(kind=rel_schema.peer, branch=self._branch)
                peer_node = InfrahubNodeSync(client=self._client, schema=peer_schema, branch=self._branch)
                peer_data = peer_node.generate_query_data_node(
                    property=property,
                    include_metadata=include_metadata,
                )

            rel_data: dict[str, Any]
            if rel_schema and rel_schema.cardinality == "one":
                rel_data = RelatedNodeSync._generate_query_data(
                    peer_data=peer_data, property=property, include_metadata=include_metadata
                )
                # Nodes involved in a hierarchy are required to inherit from a common ancestor node, and graphql
                # tries to resolve attributes in this ancestor instead of actual node. To avoid
                # invalid queries issues when attribute is missing in the common ancestor, we use a fragment
                # to explicit actual node kind we are querying.
                if rel_schema.kind == RelationshipKind.HIERARCHY:
                    data_node = rel_data["node"]
                    rel_data["node"] = {}
                    rel_data["node"][f"...on {rel_schema.peer}"] = data_node
            elif rel_schema and rel_schema.cardinality == "many":
                rel_data = RelationshipManagerSync._generate_query_data(
                    peer_data=peer_data, property=property, include_metadata=include_metadata
                )
            else:
                continue

            data[rel_name] = rel_data

            if insert_alias:
                data[rel_name]["@alias"] = f"__alias__{self._schema.kind}__{rel_name}"

        # Add parent, children, ancestors and descendants for hierarchical nodes
        self._process_hierarchical_fields(
            data=data,
            include=include,
            exclude=exclude,
            prefetch_relationships=prefetch_relationships,
            insert_alias=insert_alias,
            property=property,
        )

        return data

    def add_relationships(
        self,
        relation_to_update: str,
        related_nodes: list[str],
    ) -> None:
        query = self._relationship_mutation(
            action="Add", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipadd-{relation_to_update}"
        self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Remove", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipremove-{relation_to_update}"
        self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def _generate_mutation_query(self) -> dict[str, Any]:
        query_result: dict[str, Any] = {"ok": None, "object": {"id": None}}

        for attr_name in self._attributes:
            attr: Attribute = getattr(self, attr_name)
            query_result["object"].update(attr._generate_mutation_query())

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if not isinstance(rel, RelatedNodeSync):
                continue

            query_result["object"].update(rel._generate_mutation_query())

        return query_result

    def _process_mutation_result(
        self, mutation_name: str, response: dict[str, Any], timeout: int | None = None
    ) -> None:
        object_response: dict[str, Any] = response[mutation_name]["object"]
        self.id = object_response["id"]
        self._existing = True

        for attr_name in self._attributes:
            attr = getattr(self, attr_name)
            if attr_name not in object_response or not attr.is_from_pool_attribute():
                continue

            # Process allocated resource from a pool and update attribute
            attr.value = object_response[attr_name]["value"]

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel_name not in object_response or not isinstance(rel, RelatedNodeSync) or not rel.is_resource_pool:
                continue

            # Process allocated resource from a pool and update related node
            allocated_resource = object_response[rel_name]
            related_node = RelatedNodeSync(
                client=self._client, branch=self._branch, schema=rel.schema, data=allocated_resource
            )
            related_node.fetch(timeout=timeout)
            setattr(self, rel_name, related_node)

    def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        if self._file_object_support and self._file_content is None:
            raise ValueError(
                f"Cannot create {self._schema.kind} without file content. Use upload_from_path() or upload_from_bytes() to provide "
                "file content before saving."
            )

        mutation_query = self._generate_mutation_query()

        if allow_upsert:
            input_data = self._generate_input_data(exclude_hfid=False, request_context=request_context)
            mutation_name = f"{self._schema.kind}Upsert"
            tracker = f"mutation-{str(self._schema.kind).lower()}-upsert"
        else:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Create"
            tracker = f"mutation-{str(self._schema.kind).lower()}-create"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        if "file" in input_data["mutation_variables"]:
            prepared = self._get_file_for_upload_sync()
            try:
                response = self._client._execute_graphql_with_file(
                    query=query.render(),
                    variables=input_data["variables"],
                    file_content=prepared.file_object,
                    file_name=prepared.filename,
                    branch_name=self._branch,
                    tracker=tracker,
                    timeout=timeout,
                )
            finally:
                if prepared.should_close and prepared.file_object:
                    prepared.file_object.close()
            # Clear the file content after successful upload
            self.clear_file()
        else:
            response = self._client.execute_graphql(
                query=query.render(),
                branch_name=self._branch,
                tracker=tracker,
                variables=input_data["variables"],
                timeout=timeout,
            )
        self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    def update(
        self, do_full_update: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        input_data = self._generate_input_data(exclude_unmodified=not do_full_update, request_context=request_context)
        mutation_query = self._generate_mutation_query()
        mutation_name = f"{self._schema.kind}Update"
        tracker = f"mutation-{str(self._schema.kind).lower()}-update"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        if "file" in input_data["mutation_variables"]:
            prepared = self._get_file_for_upload_sync()
            try:
                response = self._client._execute_graphql_with_file(
                    query=query.render(),
                    variables=input_data["variables"],
                    file_content=prepared.file_object,
                    file_name=prepared.filename,
                    branch_name=self._branch,
                    tracker=tracker,
                    timeout=timeout,
                )
            finally:
                if prepared.should_close and prepared.file_object:
                    prepared.file_object.close()
            # Clear the file content after successful upload
            self.clear_file()
        else:
            response = self._client.execute_graphql(
                query=query.render(),
                branch_name=self._branch,
                tracker=tracker,
                variables=input_data["variables"],
                timeout=timeout,
            )
        self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    def _process_relationships(
        self,
        node_data: dict[str, Any],
        branch: str,
        related_nodes: list[InfrahubNodeSync],
        timeout: int | None = None,
        recursive: bool = False,
    ) -> None:
        """Processes the Relationships of a InfrahubNodeSync and add Related Nodes to a list.

        Args:
            node_data (dict[str, Any]): The item from the GraphQL response corresponding to the node.
            branch (str): The branch name.
            related_nodes (list[InfrahubNodeSync]): The list to which related nodes will be appended.
            timeout (int, optional): Overrides default timeout used when querying the graphql API. Specified in seconds.
            recursive:(bool): Whether to recursively process relationships of related nodes.
        """
        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel and isinstance(rel, RelatedNodeSync):
                relation = node_data["node"].get(rel_name, None)
                if relation and relation.get("node", None):
                    related_node = InfrahubNodeSync.from_graphql(
                        client=self._client,
                        branch=branch,
                        data=relation,
                        timeout=timeout,
                    )
                    related_nodes.append(related_node)
                    if recursive:
                        related_node._process_relationships(
                            node_data=relation,
                            branch=branch,
                            related_nodes=related_nodes,
                            recursive=recursive,
                        )
            elif rel and isinstance(rel, RelationshipManagerSync):
                peers = node_data["node"].get(rel_name, None)
                if peers and peers["edges"]:
                    for peer in peers["edges"]:
                        related_node = InfrahubNodeSync.from_graphql(
                            client=self._client,
                            branch=branch,
                            data=peer,
                            timeout=timeout,
                        )
                        related_nodes.append(related_node)
                        if recursive:
                            related_node._process_relationships(
                                node_data=peer,
                                branch=branch,
                                related_nodes=related_nodes,
                                recursive=recursive,
                            )

    def get_pool_allocated_resources(self, resource: InfrahubNodeSync) -> list[InfrahubNodeSync]:
        """Fetch all nodes that were allocated for the pool and a given resource.

        Args:
            resource (InfrahubNodeSync): The resource from which the nodes were allocated.

        Returns:
            list[InfrahubNodeSync]: The allocated nodes.
        """
        if not self.is_resource_pool():
            raise ValueError("Allocate resources can only be fetched from resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolAllocated"
        node_ids_per_kind: dict[str, list[str]] = {}

        has_remaining_items = True
        page_number = 1
        while has_remaining_items:
            page_offset = (page_number - 1) * self._client.pagination_size

            query = Query(
                query={
                    graphql_query_name: {
                        "@filters": {
                            "pool_id": "$pool_id",
                            "resource_id": "$resource_id",
                            "offset": page_offset,
                            "limit": self._client.pagination_size,
                        },
                        "count": None,
                        "edges": {"node": {"id": None, "kind": None, "branch": None, "identifier": None}},
                    }
                },
                name="GetAllocatedResourceForPool",
                variables={"pool_id": str, "resource_id": str},
            )
            response = self._client.execute_graphql(
                query=query.render(),
                variables={"pool_id": self.id, "resource_id": resource.id},
                branch_name=self._branch,
                tracker=f"get-allocated-resources-page{page_number}",
            )

            for edge in response[graphql_query_name]["edges"]:
                node = edge["node"]
                node_ids_per_kind.setdefault(node["kind"], []).append(node["id"])

            remaining_items = response[graphql_query_name].get("count", 0) - (
                page_offset + self._client.pagination_size
            )
            if remaining_items <= 0:
                has_remaining_items = False

            page_number += 1

        nodes: list[InfrahubNodeSync] = []
        for kind, node_ids in node_ids_per_kind.items():
            nodes.extend(self._client.filters(kind=kind, branch=self._branch, ids=node_ids))

        return nodes

    def get_pool_resources_utilization(self) -> list[dict[str, Any]]:
        """Fetch the utilization of each resource for the pool.

        Returns:
            list[dict[str, Any]]: A list containing the allocation numbers for each resource of the pool.
        """
        if not self.is_resource_pool():
            raise ValueError("Pool utilization can only be fetched for resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolUtilization"

        query = Query(
            query={
                graphql_query_name: {
                    "@filters": {"pool_id": "$pool_id"},
                    "count": None,
                    "edges": {
                        "node": {
                            "id": None,
                            "kind": None,
                            "utilization": None,
                            "utilization_branches": None,
                            "utilization_default_branch": None,
                        }
                    },
                }
            },
            name="GetUtilizationForPool",
            variables={"pool_id": str},
        )
        response = self._client.execute_graphql(
            query=query.render(),
            variables={"pool_id": self.id},
            branch_name=self._branch,
            tracker="get-pool-utilization",
        )

        if response[graphql_query_name].get("count", 0):
            return [edge["node"] for edge in response[graphql_query_name]["edges"]]
        return []

    def _get_relationship_many(self, name: str) -> RelationshipManager | RelationshipManagerSync:
        if name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=many relationship for {name}")

    def _get_relationship_one(self, name: str) -> RelatedNode | RelatedNodeSync:
        if name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=one relationship for {name}")

    def get_flat_value(self, key: str, separator: str = "__") -> Any:
        """Query recursively a value defined in a flat notation (string), on a hierarchy of objects

        Examples:
            name__value
            module.object.value
        """
        if separator not in key:
            return getattr(self, key)

        first, remaining = key.split(separator, maxsplit=1)

        if first in self._schema.attribute_names:
            attr = getattr(self, first)
            for part in remaining.split(separator):
                attr = getattr(attr, part)
            return attr

        try:
            rel = self._schema.get_relationship(name=first)
        except ValueError as exc:
            raise ValueError(f"No attribute or relationship named '{first}' for '{self._schema.kind}'") from exc

        if rel.cardinality != RelationshipCardinality.ONE:
            raise ValueError(
                f"Can only look up flat value for relationships of cardinality {RelationshipCardinality.ONE.value}"
            )

        related_node: RelatedNodeSync = getattr(self, first)
        related_node.fetch()
        return related_node.peer.get_flat_value(key=remaining, separator=separator)

    def extract(self, params: dict[str, str]) -> dict[str, Any]:
        """Extract some data points defined in a flat notation."""
        result: dict[str, Any] = {}
        for key, value in params.items():
            result[key] = self.get_flat_value(key=value)

        return result

    def __dir__(self) -> Iterable[str]:
        base = list(super().__dir__())
        return sorted(
            base
            + list(self._attribute_data.keys())
            + list(self._relationship_cardinality_many_data.keys())
            + list(self._relationship_cardinality_one_data.keys())
        )



import ipaddress
import re

PROPERTIES_FLAG = ["is_protected", "updated_at"]
PROPERTIES_OBJECT = ["source", "owner"]

# Attribute-level metadata object fields (in addition to PROPERTIES_OBJECT)
ATTRIBUTE_METADATA_OBJECT = ["updated_by"]

# Node metadata fields (for node_metadata in GraphQL response)
NODE_METADATA_FIELDS_FLAG = ["created_at", "updated_at"]
NODE_METADATA_FIELDS_OBJECT = ["created_by", "updated_by"]

# Relationship metadata fields (for relationship_metadata in GraphQL response)
RELATIONSHIP_METADATA_FIELDS_FLAG = ["updated_at"]
RELATIONSHIP_METADATA_FIELDS_OBJECT = ["updated_by"]
SAFE_VALUE = re.compile(r"(^[\. /:a-zA-Z0-9_-]+$)|(^$)")

IP_TYPES = ipaddress.IPv4Interface | ipaddress.IPv6Interface | ipaddress.IPv4Network | ipaddress.IPv6Network

ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_fetch is only supported for nodes that are Artifact Definition target"
)
ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling artifact_generate is only supported for nodes that are Artifact Definition targets"
)
ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling generate is only supported for CoreArtifactDefinition nodes"
)
FILE_DOWNLOAD_FEATURE_NOT_SUPPORTED_MESSAGE = (
    "calling download_file is only supported for nodes that inherit from CoreFileObject"
)

HIERARCHY_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE = "Hierarchical fields are not supported for this node."

HFID_STR_SEPARATOR = "__"
PROFILE_KIND_PREFIX = "Profile"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.constants.relationship_label import RELATIONSHIP_TO_VALUE_LABEL
from infrahub.core.graph.schema import GraphAttributeValueIndexedNode, GraphAttributeValueNode
from infrahub.core.query import Query, QueryType
from infrahub.types import is_large_attribute_type

from .model import QueryAttributePathValued, QueryRelationshipPathValued

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

    from .model import NodeUniquenessQueryRequest, NodeUniquenessQueryRequestValued


class NodeUniqueAttributeConstraintQuery(Query):
    name = "node_constraints_uniqueness"
    insert_return = False
    type = QueryType.READ
    attribute_property_map = {"value": RELATIONSHIP_TO_VALUE_LABEL}

    def __init__(
        self,
        query_request: NodeUniquenessQueryRequest,
        min_count_required: int = 1,
        **kwargs: Any,
    ) -> None:
        self.query_request = query_request
        self.min_count_required = min_count_required
        super().__init__(**kwargs)

    def get_context(self) -> dict[str, str]:
        return {"kind": self.query_request.kind}

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002,PLR0915
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)
        from_times = db.render_list_comprehension(items="relationships(potential_path)", item_name="from")
        branch_name_and_level = db.render_list_comprehension_with_list(
            items="relationships(active_path)", item_names=["branch", "branch_level"]
        )

        attrs_include_large_type = False
        attribute_names = set()
        attr_paths, attr_paths_with_value, attr_values = [], [], []
        for attr_path in self.query_request.unique_attribute_paths:
            try:
                property_rel_name = self.attribute_property_map[attr_path.property_name or "value"]
            except KeyError as exc:
                raise ValueError(
                    f"{attr_path.property_name} is not a valid property for a uniqueness constraint"
                ) from exc
            if is_large_attribute_type(attr_path.attribute_kind):
                attrs_include_large_type = True
            attribute_names.add(attr_path.attribute_name)
            if attr_path.value:
                attr_paths_with_value.append((attr_path.attribute_name, property_rel_name, attr_path.value))
                attr_values.append(attr_path.value)
            else:
                attr_paths.append((attr_path.attribute_name, property_rel_name))
        attr_value_label = (
            GraphAttributeValueNode.get_default_label()
            if attrs_include_large_type
            else GraphAttributeValueIndexedNode.get_default_label()
        )

        relationship_names = set()
        relationship_attr_paths = []
        relationship_only_attr_paths = []
        relationship_only_attr_values = []
        relationship_attr_values = []
        relationship_attr_paths_with_value = []
        for rel_path in self.query_request.relationship_attribute_paths:
            relationship_names.add(rel_path.identifier)
            if rel_path.attribute_name and rel_path.value:
                relationship_attr_paths_with_value.append(
                    (rel_path.identifier, rel_path.attribute_name, rel_path.value)
                )
                relationship_attr_values.append(rel_path.value)
            elif rel_path.attribute_name:
                relationship_attr_paths.append((rel_path.identifier, rel_path.attribute_name))
            else:
                relationship_only_attr_paths.append(rel_path.identifier)
                if rel_path.value:
                    relationship_only_attr_values.append(rel_path.value)

        if (
            not attr_paths
            and not attr_paths_with_value
            and not relationship_attr_paths
            and not relationship_attr_paths_with_value
            and not relationship_only_attr_paths
        ):
            raise ValueError(
                "The NodeUniquenessQueryRequest provided for node_constraints_uniqueness doesn't have enough information to continue"
            )

        self.params.update(
            {
                "node_kind": self.query_request.kind,
                "attr_paths": attr_paths,
                "attr_paths_with_value": attr_paths_with_value,
                "attr_values": attr_values,
                "attribute_names": list(attribute_names),
                "relationship_names": list(relationship_names),
                "relationship_attr_paths": relationship_attr_paths,
                "relationship_attr_paths_with_value": relationship_attr_paths_with_value,
                "relationship_only_attr_paths": relationship_only_attr_paths,
                "relationship_only_attr_values": relationship_only_attr_values,
                "relationship_attr_values": relationship_attr_values,
                "min_count_required": self.min_count_required,
            }
        )

        attr_paths_subquery = """
        MATCH attr_path = (start_node:%(node_kind)s)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[r:HAS_VALUE]->(attr_value:AttributeValue)
        WHERE attr.name in $attribute_names
            AND [attr.name, type(r)] in $attr_paths
        RETURN start_node, attr_path as potential_path, NULL as rel_identifier, attr.name as potential_attr, attr_value.value as potential_attr_value
        """ % {"node_kind": self.query_request.kind}

        attr_paths_with_value_subquery = """
        MATCH attr_path = (start_node:%(node_kind)s)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[r:HAS_VALUE]->(attr_value:%(attr_value_label)s)
        WHERE attr.name in $attribute_names AND attr_value.value in $attr_values
            AND [attr.name, type(r), attr_value.value] in $attr_paths_with_value
        RETURN start_node, attr_path as potential_path, NULL as rel_identifier, attr.name as potential_attr, attr_value.value as potential_attr_value
        """ % {"node_kind": self.query_request.kind, "attr_value_label": attr_value_label}

        relationship_attr_paths_subquery = """
        MATCH rel_path = (start_node:%(node_kind)s)-[:IS_RELATED]-(relationship_node:Relationship)-[:IS_RELATED]-(related_n:Node)-[:HAS_ATTRIBUTE]->(rel_attr:Attribute)-[:HAS_VALUE]->(rel_attr_value:AttributeValue)
        WHERE relationship_node.name in $relationship_names
            AND [relationship_node.name, rel_attr.name] in $relationship_attr_paths
        RETURN start_node, rel_path as potential_path, relationship_node.name as rel_identifier, rel_attr.name as potential_attr, rel_attr_value.value as potential_attr_value
        """ % {"node_kind": self.query_request.kind}

        relationship_attr_paths_with_value_subquery = """
        MATCH rel_path = (start_node:%(node_kind)s)-[:IS_RELATED]-(relationship_node:Relationship)-[:IS_RELATED]-(related_n:Node)-[:HAS_ATTRIBUTE]->(rel_attr:Attribute)-[:HAS_VALUE]->(rel_attr_value:AttributeValue)
        WHERE relationship_node.name in $relationship_names AND rel_attr_value.value in $relationship_attr_values
            AND [relationship_node.name, rel_attr.name, rel_attr_value.value] in $relationship_attr_paths_with_value
        RETURN start_node, rel_path as potential_path, relationship_node.name as rel_identifier, rel_attr.name as potential_attr, rel_attr_value.value as potential_attr_value
        """ % {"node_kind": self.query_request.kind}

        relationship_only_attr_paths_subquery = """
        MATCH rel_path = (start_node:%(node_kind)s)-[:IS_RELATED]-(relationship_node:Relationship)-[:IS_RELATED]-(related_n:Node)
        WHERE %(rel_node_filter)s relationship_node.name in $relationship_only_attr_paths
        RETURN start_node, rel_path as potential_path, relationship_node.name as rel_identifier, "id" as potential_attr, related_n.uuid as potential_attr_value
        """ % {
            "node_kind": self.query_request.kind,
            "rel_node_filter": "related_n.uuid IN $relationship_only_attr_values AND "
            if relationship_only_attr_values
            else "",
        }

        select_subqueries = []
        if attr_paths:
            select_subqueries.append(attr_paths_subquery)
        if attr_paths_with_value:
            select_subqueries.append(attr_paths_with_value_subquery)
        if relationship_attr_paths:
            select_subqueries.append(relationship_attr_paths_subquery)
        if relationship_attr_paths_with_value:
            select_subqueries.append(relationship_attr_paths_with_value_subquery)
        if relationship_only_attr_paths:
            select_subqueries.append(relationship_only_attr_paths_subquery)

        select_subqueries_str = "UNION".join(select_subqueries)

        # ruff: noqa: E501
        query = """
        // get attributes for node and its relationships
        CALL () {
            %(select_subqueries_str)s
        }
        CALL (potential_path) {
            WITH potential_path
            // only the branches and times we care about
            WHERE all(
                r IN relationships(potential_path) WHERE (
                    %(branch_filter)s
                )
            )
            // only get the latest path on the farthest branch from main
            RETURN
                potential_path as matched_path,
                reduce(br_lvl = 0, r in relationships(potential_path) | br_lvl + r.branch_level) AS branch_level_sum,
                %(from_times)s AS from_times,
                // used as tiebreaker for updated relationships that were deleted and added at the same microsecond
                reduce(active_count = 0, r in relationships(potential_path) | active_count + (CASE r.status WHEN "active" THEN 1 ELSE 0 END)) AS active_relationship_count
        }
        WITH
            collect([matched_path, branch_level_sum, from_times, active_relationship_count, potential_attr_value]) as enriched_paths,
            start_node,
            rel_identifier,
            potential_attr
        CALL (enriched_paths) {
            UNWIND enriched_paths as path_to_check
            RETURN path_to_check[0] as current_path, path_to_check[4] as latest_value
            ORDER BY
                path_to_check[1] DESC,
                path_to_check[2][-1] DESC,
                path_to_check[2][-2] DESC,
                path_to_check[3] DESC
            LIMIT 1
        }
        CALL (current_path) {
            // only active paths
            WITH current_path
            WHERE all(r IN relationships(current_path) WHERE r.status = "active")
            RETURN current_path as active_path
        }
        CALL (active_path) {
            // get deepest branch name
            UNWIND %(branch_name_and_level)s as branch_name_and_level
            RETURN branch_name_and_level[0] as branch_name
            ORDER BY branch_name_and_level[1] DESC
            LIMIT 1
        }
        // only duplicate values
        WITH
            collect([start_node.uuid, branch_name]) as nodes_and_branches,
            count(*) as node_count,
            potential_attr as attr_name,
            latest_value as attr_value,
            rel_identifier as relationship_identifier
        WHERE node_count > $min_count_required
        UNWIND nodes_and_branches as node_and_branch
        RETURN
            node_and_branch[0] as node_id,
            node_and_branch[1] as deepest_branch_name,
            node_count,
            attr_name,
            attr_value,
            relationship_identifier
        ORDER BY
            node_id,
            deepest_branch_name,
            node_count,
            attr_name,
            attr_value,
            relationship_identifier
        """ % {
            "select_subqueries_str": select_subqueries_str,
            "branch_filter": branch_filter,
            "from_times": from_times,
            "branch_name_and_level": branch_name_and_level,
        }

        self.add_to_query(query)
        self.return_labels = [
            "node_id",
            "deepest_branch_name",
            "node_count",
            "attr_name",
            "attr_value",
            "relationship_identifier",
        ]


class UniquenessValidationQuery(Query):
    name = "uniqueness_constraint_validation"
    type = QueryType.READ

    def __init__(
        self,
        query_request: NodeUniquenessQueryRequestValued,
        node_ids_to_exclude: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        self.query_request = query_request
        self.node_ids_to_exclude = node_ids_to_exclude
        super().__init__(**kwargs)

    def _is_attribute_large_type(self, db: InfrahubDatabase, node_kind: str, attribute_name: str) -> bool:
        """Determine if an attribute is a large type that should use AttributeValue instead of AttributeValueIndexed."""
        node_schema = db.schema.get(node_kind, branch=self.branch, duplicate=False)
        attr_schema = node_schema.get_attribute(attribute_name)
        return is_large_attribute_type(attr_schema.kind)

    def _build_attr_subquery(
        self,
        node_kind: str,
        attr_path: QueryAttributePathValued,
        index: int,
        branch_filter: str,
        is_first_query: bool,
        is_large_type: bool,
    ) -> tuple[str, dict[str, str | int | float | bool]]:
        attr_name_var = f"attr_name_{index}"
        attr_value_var = f"attr_value_{index}"
        if is_first_query:
            first_query_filter = "WHERE $node_ids_to_exclude IS NULL OR NOT node.uuid IN $node_ids_to_exclude"
        else:
            first_query_filter = ""

        # Determine the appropriate label based on attribute type
        attr_value_label = (
            GraphAttributeValueNode.get_default_label()
            if is_large_type
            else GraphAttributeValueIndexedNode.get_default_label()
        )

        attribute_query = """
MATCH (node:%(node_kind)s)-[:HAS_ATTRIBUTE]->(attr:Attribute {name: $%(attr_name_var)s})-[:HAS_VALUE]->(:%(attr_value_label)s {value: $%(attr_value_var)s})
%(first_query_filter)s
WITH DISTINCT node
CALL (node) {
    MATCH (node)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $%(attr_name_var)s})
    WHERE %(branch_filter)s
    WITH attr, r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    WITH attr, is_active
    LIMIT 1
    WITH attr, is_active
    WHERE is_active = TRUE
    MATCH (attr)-[r:HAS_VALUE]->(:AttributeValue {value: $%(attr_value_var)s})
    WHERE %(branch_filter)s
    WITH r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH r
    WHERE r.status = "active"
    RETURN 1 AS is_match_%(index)s
}
        """ % {
            "first_query_filter": first_query_filter,
            "node_kind": node_kind,
            "attr_name_var": attr_name_var,
            "attr_value_var": attr_value_var,
            "branch_filter": branch_filter,
            "index": index,
            "attr_value_label": attr_value_label,
        }
        params: dict[str, str | int | float | bool] = {
            attr_name_var: attr_path.attribute_name,
            attr_value_var: attr_path.value,
        }
        return attribute_query, params

    def _build_rel_subquery(
        self,
        node_kind: str,
        rel_path: QueryRelationshipPathValued,
        index: int,
        branch_filter: str,
        is_first_query: bool,
        is_large_type: bool = False,
    ) -> tuple[str, dict[str, str | int | float | bool]]:
        params: dict[str, str | int | float | bool] = {}
        rel_attr_query = ""
        rel_attr_match = ""
        if rel_path.attribute_name and rel_path.attribute_value:
            attr_name_var = f"attr_name_{index}"
            attr_value_var = f"attr_value_{index}"

            # Determine the appropriate label based on relationship attribute type
            rel_attr_value_label = (
                GraphAttributeValueNode.get_default_label()
                if is_large_type
                else GraphAttributeValueIndexedNode.get_default_label()
            )

            rel_attr_query = """
    MATCH (peer)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $%(attr_name_var)s})
    WHERE %(branch_filter)s
    WITH attr, r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    WITH attr, is_active
    LIMIT 1
    WITH attr, is_active
    WHERE is_active = TRUE
    MATCH (attr)-[r:HAS_VALUE]->(:%(rel_attr_value_label)s {value: $%(attr_value_var)s})
    WHERE %(branch_filter)s
    WITH r
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH r
    WHERE r.status = "active"
            """ % {
                "attr_name_var": attr_name_var,
                "attr_value_var": attr_value_var,
                "branch_filter": branch_filter,
                "rel_attr_value_label": rel_attr_value_label,
            }
            rel_attr_match = (
                "-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $%(attr_name_var)s})-[:HAS_VALUE]->(:%(rel_attr_value_label)s {value: $%(attr_value_var)s})"
                % {
                    "attr_name_var": attr_name_var,
                    "attr_value_var": attr_value_var,
                    "rel_attr_value_label": rel_attr_value_label,
                }
            )
            params[attr_name_var] = rel_path.attribute_name
            params[attr_value_var] = rel_path.attribute_value
        query_arrows = rel_path.relationship_schema.get_query_arrows()
        rel_name_var = f"rel_name_{index}"
        # long path MATCH is required to hit an index on the peer or AttributeValue of the peer
        first_match = (
            "MATCH (node:%(node_kind)s)%(lstart)s[:IS_RELATED]%(lend)s(:Relationship {name: $%(rel_name_var)s})%(rstart)s[:IS_RELATED]%(rend)s"
            % {
                "node_kind": node_kind,
                "lstart": query_arrows.left.start,
                "lend": query_arrows.left.end,
                "rstart": query_arrows.right.start,
                "rend": query_arrows.right.end,
                "rel_name_var": rel_name_var,
            }
        )
        peer_where = f"WHERE {branch_filter}"
        if rel_path.peer_id:
            peer_id_var = f"peer_id_{index}"
            peer_where += f" AND peer.uuid = ${peer_id_var}"
            params[peer_id_var] = rel_path.peer_id
            first_match += "(:Node {uuid: $%(peer_id_var)s})" % {"peer_id_var": peer_id_var}
        else:
            peer_where += " AND peer.uuid <> node.uuid"
            first_match += "(:Node)"
        if rel_attr_match:
            first_match += rel_attr_match
        if is_first_query:
            first_query_filter = "WHERE $node_ids_to_exclude IS NULL OR NOT node.uuid IN $node_ids_to_exclude"
        else:
            first_query_filter = ""
        relationship_query = f"""
{first_match}
{first_query_filter}
WITH DISTINCT node
        """
        relationship_query += """
CALL (node) {
    MATCH (node)%(lstart)s[r:IS_RELATED]%(lend)s(rel:Relationship {name: $%(rel_name_var)s})
    WHERE %(branch_filter)s
    WITH rel, r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    WITH rel, is_active
    LIMIT 1
    WITH rel, is_active
    WHERE is_active = TRUE
    MATCH (rel)%(rstart)s[r:IS_RELATED]%(rend)s(peer:Node)
    %(peer_where)s
    WITH peer, r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    WITH peer, is_active
    LIMIT 1
    WITH peer, is_active
    WHERE is_active = TRUE
%(rel_attr_query)s
    RETURN 1 AS is_match_%(index)s
    LIMIT 1
}
        """ % {
            "rel_name_var": rel_name_var,
            "lstart": query_arrows.left.start,
            "lend": query_arrows.left.end,
            "rstart": query_arrows.right.start,
            "rend": query_arrows.right.end,
            "peer_where": peer_where,
            "rel_attr_query": rel_attr_query,
            "branch_filter": branch_filter,
            "index": index,
        }
        params[rel_name_var] = rel_path.relationship_schema.get_identifier()
        return relationship_query, params

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params["node_ids_to_exclude"] = self.node_ids_to_exclude
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at.to_string(), is_isolated=False)
        self.params.update(branch_params)

        subqueries = []
        for index, schema_path in enumerate(self.query_request.unique_valued_paths):
            is_first_query = index == 0
            if isinstance(schema_path, QueryAttributePathValued):
                is_large_type = self._is_attribute_large_type(
                    db=db, node_kind=self.query_request.kind, attribute_name=schema_path.attribute_name
                )
                subquery, params = self._build_attr_subquery(
                    node_kind=self.query_request.kind,
                    attr_path=schema_path,
                    index=index,
                    branch_filter=branch_filter,
                    is_first_query=is_first_query,
                    is_large_type=is_large_type,
                )
            else:
                subquery, params = self._build_rel_subquery(
                    node_kind=self.query_request.kind,
                    rel_path=schema_path,
                    index=index,
                    branch_filter=branch_filter,
                    is_first_query=is_first_query,
                )
            subqueries.append(subquery)
            self.params.update(params)

        full_query = "\n".join(subqueries)
        self.add_to_query(full_query)
        self.return_labels = ["node.uuid AS node_uuid", "node.kind AS node_kind"]

    def get_violation_nodes(self) -> list[tuple[str, str]]:
        violation_tuples = []
        for result in self.results:
            violation_tuples.append(
                (result.get_as_type("node_uuid", return_type=str), result.get_as_type("node_kind", return_type=str))
            )
        return violation_tuples



from typing import Any

from prefect import task
from prefect.cache_policies import NONE
from prefect.client.orchestration import PrefectClient, get_client
from prefect.client.schemas.objects import WorkerStatus

from infrahub.events.utils import get_all_events
from infrahub.trigger.constants import NAME_SEPARATOR
from infrahub.trigger.models import TriggerType
from infrahub.trigger.setup import gather_all_automations

from .models import TelemetryPrefectData, TelemetryWorkPoolData


@task(name="telemetry-gather-work-pools", task_run_name="Gather Work Pools", cache_policy=NONE)
async def gather_prefect_work_pools(client: PrefectClient) -> list[TelemetryWorkPoolData]:
    work_pools = await client.read_work_pools()
    data: list[TelemetryWorkPoolData] = []

    for pool in work_pools:
        workers = await client.read_workers_for_work_pool(work_pool_name=pool.name)
        data.append(
            TelemetryWorkPoolData(
                name=pool.name,
                type=pool.type,
                total_workers=len(workers),
                active_workers=len([item for item in workers if item.status == WorkerStatus.ONLINE]),
            )
        )

    return data


@task(name="telemetry-gather-events", task_run_name="Gather Events", cache_policy=NONE)
async def gather_prefect_events(client: PrefectClient) -> dict[str, Any]:
    infrahub_events = get_all_events()
    events: dict[str, int] = {}

    async def count_events(event_name: str) -> int:
        payload = {"filter": {"event": {"name": [event_name]}}}
        response = await client._client.post("/events/count-by/event", json=payload)
        response.raise_for_status()
        data = response.json()
        if not isinstance(data, list) or len(data) == 0:
            return 0
        return data[0]["count"]

    for event in infrahub_events:
        events[event.event_name] = await count_events(event_name=event.event_name)

    return events


@task(name="telemetry-gather-automations", task_run_name="Gather Automations", cache_policy=NONE)
async def gather_prefect_automations(client: PrefectClient) -> dict[str, Any]:
    automations = await gather_all_automations(client=client)

    data: dict[str, Any] = {}

    for trigger_type in TriggerType:
        data[trigger_type.value] = len(
            [item for item in automations if item.name.startswith(f"{trigger_type.value}{NAME_SEPARATOR}")]
        )

    return data


@task(name="telemetry-gather-prefect-information", task_run_name="Gather Prefect Information", cache_policy=NONE)
async def gather_prefect_information() -> TelemetryPrefectData:
    async with get_client(sync_client=False) as client:
        data = TelemetryPrefectData(
            work_pools=await gather_prefect_work_pools(client=client),
            events=await gather_prefect_events(client=client),
            automations=await gather_prefect_automations(client=client),
        )

        return data



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.schema import NodeSchema, ProfileSchema, TemplateSchema

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase


def find_node_schema(
    db: InfrahubDatabase, branch: Branch | str, labels: list[str], duplicate: bool = False
) -> NodeSchema | ProfileSchema | TemplateSchema | None:
    for label in labels:
        if db.schema.has(name=label, branch=branch):
            schema = db.schema.get(name=label, branch=branch, duplicate=duplicate)
            if isinstance(schema, NodeSchema | ProfileSchema | TemplateSchema):
                return schema

    return None


def filter_and(items: list[str]) -> str:
    filter_str = " AND ".join(items)
    if len(items) > 1:
        return f" ( {filter_str} ) "
    return filter_str


def filter_or(items: list[str]) -> str:
    filter_str = " OR ".join(items)
    if len(items) > 1:
        return f" ( {filter_str} ) "
    return filter_str



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Mapping

from infrahub.core.constants import RelationshipCardinality
from infrahub.exceptions import ValidationError

from .interface import RelationshipManagerConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.schema import MainSchemaTypes, NonGenericSchemaTypes
    from infrahub.database import InfrahubDatabase

    from ..model import RelationshipManager


@dataclass
class NodeToValidate:
    uuid: str
    relative_uuids: set[str]
    schema: NonGenericSchemaTypes


class RelationshipPeerRelativesConstraint(RelationshipManagerConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    async def _check_relationship_peers_relatives(
        self,
        relm: RelationshipManager,
        node_schema: MainSchemaTypes,
        peers: Mapping[str, Node],
        relationship_name: str,
    ) -> None:
        """Validate that all peers of a given `relm` have the same set of relatives (aka peers) for the given `relationship_name`."""
        nodes_to_validate: list[NodeToValidate] = []

        for peer in peers.values():
            peer_schema = peer.get_schema()
            peer_relm: RelationshipManager = getattr(peer, relationship_name)
            peer_relm_peers = await peer_relm.get_peers(db=self.db)

            nodes_to_validate.append(
                NodeToValidate(
                    uuid=peer.id, relative_uuids={n.id for n in peer_relm_peers.values()}, schema=peer_schema
                )
            )

        relative_uuids = nodes_to_validate[0].relative_uuids
        for node in nodes_to_validate[1:]:
            if node.relative_uuids != relative_uuids:
                raise ValidationError(
                    f"All the elements of the '{relm.name}' relationship on node {node.uuid} ({node_schema.kind}) must have the same set of peers "
                    f"for their '{node.schema.kind}.{relationship_name}' relationship"
                )

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes, node: Node) -> None:  # noqa: ARG002
        if relm.schema.cardinality != RelationshipCardinality.MANY or not relm.schema.common_relatives:
            return

        peers = await relm.get_peers(db=self.db)
        if not peers:
            return

        for rel_name in relm.schema.common_relatives:
            await self._check_relationship_peers_relatives(
                relm=relm, node_schema=node_schema, peers=peers, relationship_name=rel_name
            )



from __future__ import annotations

from infrahub.constants.database import EntityType, IndexType

from .index import IndexInfo, IndexItem, IndexManagerBase


class IndexNodeMemgraph(IndexItem):
    def get_add_query(self) -> str:
        properties_str = ", ".join(self.properties)
        return f"CREATE INDEX ON :{self.label}({properties_str})"

    def get_drop_query(self) -> str:
        properties_str = ", ".join(self.properties)
        return f"DROP INDEX ON :{self.label}({properties_str})"


class IndexManagerMemgraph(IndexManagerBase):
    def init(self, nodes: list[IndexItem], rels: list[IndexItem]) -> None:  # noqa: ARG002
        self.nodes = [IndexNodeMemgraph(**item.model_dump()) for item in nodes]
        self.initialized = True

    async def add(self) -> None:
        for item in self.items:
            await self.db.execute_query(query=item.get_add_query(), params={}, name="index_add")

    async def drop(self) -> None:
        for item in self.items:
            await self.db.execute_query(query=item.get_drop_query(), params={}, name="index_drop")

    async def list(self) -> list[IndexInfo]:
        query = "SHOW INDEX INFO"
        records = await self.db.execute_query(query=query, params={}, name="index_show")
        results = []
        for record in records:
            if not record["label"]:
                continue
            results.append(
                IndexInfo(
                    name="n/a",
                    label=record["label"],
                    properties=[record["property"]],
                    type=IndexType.NOT_APPLICABLE,
                    entity_type=EntityType.NODE,  # Memgraph only support Node Indexes
                )
            )

        return results



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class RollbackQuery(Query):
    """Rollback all database changes made at a specific timestamp on a branch.

    This query reverses database changes by:
    1. Resetting `to` times back to NULL for relationships that were closed at the given timestamp
    2. Deleting relationships that were created at the given timestamp
    3. Deleting any vertices that become orphaned after the edge deletions
    """

    name = "rollback"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        at: Timestamp,
        target_branch: Branch,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.rollback_at = at
        self.target_branch = target_branch

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {
            "at": self.rollback_at.to_string(),
            "target_branch": self.target_branch.name,
        }
        query = """
// ---------------------------
// Reset 'to' times: restore relationships that were closed at this timestamp
// ---------------------------
CALL () {
    OPTIONAL MATCH ()-[r_to {to: $at, branch: $target_branch}]-()
    SET r_to.to = NULL
} IN TRANSACTIONS

// ---------------------------
// Delete 'from' times: remove relationships that were created at this timestamp
// and collect the vertex IDs of connected nodes
// ---------------------------
MATCH (s)-[r {from: $at, branch: $target_branch}]->(d)
CALL (r) {
    DELETE r
} IN TRANSACTIONS OF 500 ROWS

// ---------------------------
// Collect the database IDs of all vertices that were connected to deleted edges
// ---------------------------
WITH DISTINCT elementId(s) AS s_id, elementId(d) AS d_id
WITH collect(s_id) + collect(d_id) AS vertex_ids

// ---------------------------
// Delete any vertices that are now orphaned (have no remaining connections)
// ---------------------------
MATCH (n)
WHERE elementId(n) IN vertex_ids
AND NOT exists((n)--())
CALL (n) {
    DELETE n
} IN TRANSACTIONS OF 500 ROWS
        """
        self.add_to_query(query=query)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.initialization import initialization
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import ArbitraryMigration, MigrationInput, MigrationResult
from infrahub.core.timestamp import Timestamp
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class FindOrphanedSchemaFieldsQuery(Query):
    name = "find_orphaned_schema_fields"
    type = QueryType.WRITE

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// ------------
// Find orphaned SchemaRelationship and SchemaAttribute vertices
// ------------
MATCH (schema_field:SchemaRelationship|SchemaAttribute)-[e:IS_RELATED]-(rel:Relationship)
WHERE rel.name IN ["schema__node__relationships", "schema__node__attributes"]
AND e.status = "deleted" OR e.to IS NOT NULL
WITH schema_field, e.branch AS branch, CASE
    WHEN e.status = "deleted" THEN e.from
    ELSE e.to
END AS delete_time
CALL (schema_field, branch) {
    OPTIONAL MATCH (schema_field)-[is_part_of:IS_PART_OF {branch: branch}]->(:Root)
    WHERE is_part_of.status = "deleted" OR is_part_of.to IS NOT NULL
    RETURN is_part_of IS NOT NULL AS is_deleted
}
WITH schema_field, branch, delete_time
WHERE is_deleted = FALSE
        """
        self.add_to_query(query)
        self.return_labels = ["schema_field.uuid AS schema_field_uuid", "branch", "delete_time"]


class Migration034(ArbitraryMigration):
    """
    Finds active SchemaRelationship and SchemaAttribute vertices with deleted relationships to SchemaNodes or
    SchemaGenerics and deletes them on the same branch at the same time
    """

    name: str = "034_find_orphaned_schema_fields"
    minimum_version: int = 33
    queries: Sequence[type[Query]] = [FindOrphanedSchemaFieldsQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        db = migration_input.db
        try:
            initialize_lock()
            await initialization(db=db)
            query = await FindOrphanedSchemaFieldsQuery.init(db=db)
            await query.execute(db=db)
            schema_field_uuids_by_branch: dict[str, dict[str, str]] = defaultdict(dict)
            for result in query.get_results():
                schema_field_uuid = result.get_as_type("schema_field_uuid", return_type=str)
                branch = result.get_as_type("branch", return_type=str)
                delete_time = result.get_as_type("delete_time", return_type=str)
                schema_field_uuids_by_branch[branch][schema_field_uuid] = delete_time

            for branch, schema_rel_details in schema_field_uuids_by_branch.items():
                node_map = await NodeManager.get_many(db=db, branch=branch, ids=list(schema_rel_details.keys()))
                for schema_field_uuid, orphan_schema_rel_node in node_map.items():
                    delete_time = Timestamp(schema_rel_details[schema_field_uuid])
                    await orphan_schema_rel_node.delete(db=db, at=delete_time)
        except Exception as exc:
            log.exception("Error during orphaned schema field cleanup")
            return MigrationResult(errors=[str(exc)])

        return MigrationResult()



from typing import Any

from infrahub import config
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase


class DeleteAfterTimeQuery(Query):
    name: str = "delete_after_time"
    insert_return: bool = False
    type: QueryType = QueryType.WRITE

    def __init__(self, timestamp: Timestamp, **kwargs: Any) -> None:
        self.timestamp = timestamp
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {"timestamp": self.timestamp.to_string()}
        query_1 = """
        // ---------------------
        // Reset edges with to time after timestamp
        // ---------------------
        CALL () {
            OPTIONAL MATCH (p)-[r]-(q)
            WHERE r.to > $timestamp
            SET r.to = NULL
        }
        """
        self.add_to_query(query_1)
        if config.SETTINGS.database.db_type is config.DatabaseType.NEO4J:
            query_2 = """
            // ---------------------
            // Delete edges with from time after timestamp timestamp
            // ---------------------
            CALL () {
                OPTIONAL MATCH (p)-[r]->(q)
                WHERE r.from > $timestamp
                DELETE r
                WITH p, q
                UNWIND [p, q] AS maybe_orphan
                WITH maybe_orphan
                WHERE NOT exists((maybe_orphan)--())
                DELETE maybe_orphan
            }
            """
        else:
            query_2 = """
            // ---------------------
            // Delete edges with from time after timestamp timestamp
            // ---------------------
            CALL () {
                OPTIONAL MATCH (p)-[r]->(q)
                WHERE r.from > $timestamp
                DELETE r
            }
            """
        self.add_to_query(query_2)



from __future__ import annotations

import json
import re
from typing import TYPE_CHECKING, cast

from infrahub.exceptions import ResourceNotFoundError
from infrahub.message_bus.types import KVTTL
from infrahub.workers.dependencies import get_cache

if TYPE_CHECKING:
    from uuid import UUID

    from infrahub_sdk.diff import NodeDiff

    from infrahub.services.adapters.cache import InfrahubCache

SCHEMA_CHANGE = re.compile(r"^Schema[A-Z]")


def has_data_changes(diff_summary: list[NodeDiff], branch: str) -> bool:
    """Indicates if there are node or schema changes within the branch."""
    return any(entry["branch"] == branch for entry in diff_summary)


def has_node_changes(diff_summary: list[NodeDiff], branch: str) -> bool:
    """Indicates if there is at least one node object that has been modified in the branch"""
    return any(entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"]) for entry in diff_summary)


def get_modified_kinds(diff_summary: list[NodeDiff], branch: str) -> list[str]:
    """Return a list of non schema kinds that have been modified on the branch"""
    return list(
        {
            entry["kind"]
            for entry in diff_summary
            if entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"])
        }
    )


def get_modified_node_ids(diff_summary: list[NodeDiff], branch: str) -> list[str]:
    """Return a list of non schema nodes that have been modified on the branch"""
    return [
        entry["id"] for entry in diff_summary if entry["branch"] == branch and not SCHEMA_CHANGE.match(entry["kind"])
    ]


async def set_diff_summary_cache(pipeline_id: UUID, diff_summary: list[NodeDiff], cache: InfrahubCache) -> None:
    serialized = json.dumps(diff_summary)
    await cache.set(
        key=f"proposed_change:pipeline:pipeline_id:{pipeline_id}:diff_summary",
        value=serialized,
        expires=KVTTL.TWO_HOURS,
    )


async def get_diff_summary_cache(pipeline_id: UUID) -> list[NodeDiff]:
    cache = await get_cache()
    summary_payload = await cache.get(
        key=f"proposed_change:pipeline:pipeline_id:{pipeline_id}:diff_summary",
    )

    if not summary_payload:
        raise ResourceNotFoundError(message=f"Diff summary for pipeline {pipeline_id} was not found in the cache")

    return cast("list[NodeDiff]", json.loads(summary_payload))
